/*
**  gui.cc
**
**  mg
**
**   7th Jun 1995, adapted from ntd's gui.cc file
**  28th May 1996 - modified to use text widgets for the hitlist
**                  and rels pool windows instead of listboxes.
**
**   7th Jul 1996 - modified to use a text widget for:
**
**                  a) current query window.
**                  b) removed_terms window.
**
*/


#define _XXX_MODULE

#define G_EXTERN extern

#define OKAPI_MODULE

#include "inx-defs.h"
#include "GLOB.h"

#define ENABLE_SEARCH 1
#define DISABLE_SEARCH 0

#define ENABLE_OPTIMISE 1
#define DISABLE_OPTIMISE 0

#define DATABASE_NAME_SIZE 32

#define HITLIST_WINDOW 0
#define RELSPOOL_WINDOW 1

/*
**  'C' functions linked with TCL procedures
*/

extern int convert_runtime(ClientData clientData, Tcl_Interp *interp, 
													 int argc, char *argv[]);
extern int create_indexes(ClientData clientData, Tcl_Interp *interp, 
													int argc, char *argv[]);
extern int quit(ClientData clientData, Tcl_Interp *interp, 
										int argc, char *argv[]);
extern int set_db_type(ClientData clientData, Tcl_Interp *interp, 
											 int argc, char *argv[]);
extern int term_entry(ClientData clientData, Tcl_Interp *interp, 
											int argc, char *argv[]);
extern int window_title(ClientData clientData, Tcl_Interp *interp, 
												int argc, char *argv[]);
/*
**  Global variables for this program
*/

static int query_list_count;


int OKAPI_Init()
{
	
  /*
  **  tcl -> c  function mappings
  **
  **  c.<name> is a tcl pointer to the C function <name>
  **
  **  theInterp             tcl interpreter
	**  convert_runtime       c.convert_runtime
	**  create_indexes        c.create_indexes
  **  quit                  c.quit
  **  term_entry            c.term_entry
  **  window_title          c.window_title
  */
  
  Tcl_CreateCommand(theInterp,
										"c.convert_runtime", convert_runtime,
										(ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
	
  Tcl_CreateCommand(theInterp,
										"c.create_indexes", create_indexes,
										(ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
	
  Tcl_CreateCommand(theInterp,
										"c.quit", quit,
										(ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
	
  Tcl_CreateCommand(theInterp,
										"c.set_db_type", set_db_type,
										(ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
	
  Tcl_CreateCommand(theInterp,
										"c.term_entry", term_entry,
										(ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
	
  Tcl_CreateCommand(theInterp,
										"c.window_title", window_title,
										(ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
  
  return TCL_OK;
}


int OKAPI_Exit()
{

  /*
	**  clear up before exit
	*/
  
  Tcl_DeleteCommand(theInterp, "c.convert_runtime");
  Tcl_DeleteCommand(theInterp, "c.create_indexes");
  Tcl_DeleteCommand(theInterp, "c.quit");
  Tcl_DeleteCommand(theInterp, "c.set_db_type");
  Tcl_DeleteCommand(theInterp, "c.term_entry");
  Tcl_DeleteCommand(theInterp, "c.window_title");

  return TCL_OK;

}


/*
**  TCL Link functions
**
**
**  clear_query()
*/


/*
**  'convert_runtime'
*/

int convert_runtime(ClientData clientData, Tcl_Interp *interp, 
										int argc, char *argv[]) {
	/*
	**  argv[1] = program name
	**  argv[2] = command switch (-c)
	**  argv[3] = bss_parmpath
	**  argv[4] = db_name
	**  argv[5] = source file name
	*/

	int p;

	int return_code;
	char os_command[SMALL_BUFFER];

	fprintf(stderr, "ARGC = [%d]\n", argc);

	for (p = 1; p < argc; p++) {
		printf("argv[%d] = [%s]\n", p, argv[p]);
	}

	if ( argc != 5 ) {
		sprintf(interp->result, "ERROR");
	}						
	else {
		for (p = 1; p < argc; p++) {
			if (p == 1) {
				sprintf(os_command, "%s -c", argv[p]);
			}
			else if (p == 4) {
				sprintf(os_command + strlen(os_command), " < %s", argv[p]);
			}
			else {
				sprintf(os_command + strlen(os_command), " %s", argv[p]);
			}
		}

		fprintf(stderr, "os_command = [%s]\n", os_command);

		if (system(os_command) < 0) {
			perror(os_command);
			sprintf(interp->result, "ERROR");
		}
		else {
			sprintf(interp->result, "OK");
		}
	}
}


int create_indexes(ClientData clientData, Tcl_Interp *interp, 
									 int argc, char *argv[]) {
	/*
	**  argv[1] = "$installation_root"
	**  argv[2] = "$bss_parmpath"
	**  argv[3] = "$db_name"
	**  argv[4] = "$ni"
	*/

	int no_indexes;
	int next_index;

	char env_variable[SMALL_BUFFER];
	char env_value[SMALL_BUFFER];

	char os_command[SMALL_BUFFER];
	
	fprintf(stderr, "ARGC = [%d]\n", argc);

	if ( argc != 5 ) {
		sprintf(interp->result, "ERROR");
	}						
	else {
		sprintf(env_variable, "BSS_PARMPATH");
		sprintf(env_value, "%s", argv[2]);

		fprintf(stderr, "set_env(%s, %s)\n", env_variable, env_value);

		set_env(env_variable, env_value);

		no_indexes = atoi(argv[4]);

		sprintf(interp->result, "");

		for (next_index = 0; next_index < no_indexes; next_index++) {
			sprintf(os_command,
							"%s/bin/ix1 -c %s -doclens %s %d | %s/bin/ixf -c %s %s %d",
							argv[1], argv[2], argv[3], next_index,
							argv[1], argv[2], argv[3], next_index);

			fprintf(stderr, "os_command = [%s]\n", os_command);

			if (system(os_command) < 0) {
				perror(os_command);
				sprintf(interp->result + strlen(interp->result),
								"ERROR creating index %d", next_index);
			}
			else {
				sprintf(interp->result + strlen(interp->result),
								"Index %d CREATED OK", next_index);
			}
		}
	}
}



int quit(ClientData clientData, Tcl_Interp *interp,
						 int argc, char *argv[])
{
  int no_terms;
  int no_removed_terms;

  int next_rel;
  int term_count = 0;
	
  char temp_file[SMALL_BUFFER];
  char temp[SMALL_BUFFER];
	
  char *elapsed;
	
	/*
  RemoveTempfiles(getpid());
  ShutDown();
  ShowTime("Quit_Gui", FINISH);
	*/

  exit(0);
}




int set_db_type(ClientData clientData, Tcl_Interp *interp, 
								int argc, char *argv[])
{
	int debugging = atoi(getenv("SET_DB_TYPE_DEBUG"));

  int i;
	
	int buffer_size;
	
  char text[SMALL_BUFFER];
	
	if (argc == 2) {
		if (debugging) {
			fprintf(stderr, "gui.cc() set_db_type() -- DB_TYPE = [%s]\n", argv[1]);
		}
		sprintf(text, "OK");
	}
	else {
		sprintf(text, "ERROR");
	}

	sprintf(interp->result, "%s", text);
}


int term_entry(ClientData clientData, Tcl_Interp *interp, 
							 int argc, char *argv[])
{
  int i;
	
	int buffer_size;

  char text[SMALL_BUFFER];

	if (argc == 2) {
		sprintf(text, "%s", argv[1]);
	}
	else {
		sprintf(text, "error");
	}

	sprintf(interp->result, "%s", text);
}


int window_title(ClientData clientData, Tcl_Interp *interp, 
								 int argc, char *argv[])
{
  char dots[40];
	
  sprintf(dots, "........................................");
	
	/*
  sprintf(interp->result, "Current logfile: [%s/%s]", LOGS_DIR, TOPIC_NAME);
	*/
}


