/*
**  tcl_init.cc   --   Adapted from ntd's file
*/

#define _ENQINIT_MODULE

#define G_EXTERN

#include "inx-defs.h"
#include "GLOB.h"

/*
**  OKAPI_TclInit
**  =============
**
**  This procedure performs application-specific initialization.
**  Most applications, especially those that incorporate additional
**  packages, will have their own version of this procedure.
**
**  Results:  Returns a standard Tcl completion code, and leaves
**            an error message in interp->result if an error occurs.
**
**  Side effects:  Depends on the startup script.
*/

int OKAPI_TclInit(Tcl_Interp *interp)
{
  Tk_Window main;
  char temp[SMALL_BUFFER];

  main = Tk_MainWindow(interp);
  
  /*
  **  Call the init procedures for included packages.
  **  Each call should look like this:
  **
  **  if (Mod_Init(interp) == TCL_ERROR) {
  **    return TCL_ERROR;
  **   }
  **
  **   where "Mod" is the name of the module.
  */

  if (Tcl_Init(interp) == TCL_ERROR) {
    return TCL_ERROR;
  }

  if (Tk_Init(interp) == TCL_ERROR) {
    return TCL_ERROR;
  }

  /*
  **  Call Tcl_CreateCommand for application-specific commands,
  **  if they weren't already created by the init procedures
  **  called above.
  */

  if (OKAPI_Init() == TCL_ERROR) {
    Do_AppDeConstruct();
    return TCL_ERROR;
  }

  /*
  **  Tcl startup script:
	**
	**     ~/.okapi_rc
  */


  sprintf(tcl_RcFile, "%s/.indexing_rc", getenv("GUI_CONFIG_FILES"));

  return TCL_OK;
}


int Do_AppDeConstruct()
{
  /*
	**  DeConstruct the Application 
	*/
	
  printf("Okapi Trec 6 Interface Terminated.\n");
  OKAPI_Exit();
  exit(1);
}
