/*
**  'define.c'
**
**  mg, CISR, April 1998
*/

#define G_EXTERN extern

#include "defs.h"

/*
**  The parameter 'query_terms' is passed to the function by the tcl
**  global variable 'query' from the tcl procedure term_entry{}.
**  This will be parsed to remove spaces from the front and end and to remove
**  the adjacency operator (+)
*/

void parse_user_input(struct user_input *);
int check_for_duplicates(char *, char, int *);

G_EXTERN int calc_little_r(int);
G_EXTERN double calc_wgt(int, int, char);
G_EXTERN double calc_rsv(double, int, char);
G_EXTERN void tclsh_error(char *);
G_EXTERN void hf_header(int, char *, int);
G_EXTERN void terminate_run(char *, char *, int);
G_EXTERN void	check_against_rels(int);
G_EXTERN void malloc_error(char *, char *, char *, int);
G_EXTERN void malloc_success(char *, char *, char *, int);
G_EXTERN void current_time(char *);
G_EXTERN void show_time(char *, int);
G_EXTERN void write_termset_file(struct query_terms *, int);

void term_input(char *query_terms, char *not_found_list) {

	int debugging = atoi(getenv("TERM_INPUT_DEBUG"));

	int buffer_size;

  int existing_term_no = NONE_ASSIGNED;
  int bss_return_code;
  int chars_read;
	int changed = FALSE;
  int current_term;
  int length;
  int no_terms = 0;
  int npostings;
  int set;

  /*
	**  Terms are given correct weights/rsvs in this function. They will,
	**  however, need to be sorted into descending "rsv" order before display.
	**
	**  Terms that have np=0 are added to the string 'not_found_list'. If
	**  the length of "not_found_list" > 0 at the end it will be displayed
	**  in a pop-up window and the items in the term entry box will not be
	**  cleared.
	**
  **  The termset will be altered if np > 0 for the entered term.
  **  In this case set NEW_ITERATION to TRUE otherwise leave it alone.
  */

  struct user_input *user_query;

	int return_code;

	char bss_command[SMALL_BUFFER];
	char bss_result[SMALL_BUFFER];
  char *result_buffer;

	char *error_message;

  show_time("term_input", START);

	/*
	**  Empty "not_found_list"
	*/

	sprintf(not_found_list, "\0");

  /*
  **  Copy ""query_terms"" into "user_input.source_terms"
  */

	if (strlen(query_terms) >= MAX_TERM_LENGTH) {
		/*
		**  Too many terms entered at once.
		**  Send error message and hint back to i/f.
		*/

		buffer_size = ERROR_MESSAGE_SIZE;

		if ((error_message = (char *) malloc(buffer_size)) == NULL) {
			malloc_error("define.c", "term_input", "error_message", buffer_size);
		}

		sprintf(error_message,
						"*      The list of terms is too long. Enter them      *\n\
*              in more, smaller batches               *");
		tclsh_error(error_message);
		free(error_message);

		sprintf(not_found_list, "TOO LONG");

		hf_header(COMMAND_NO, "define", PARAMS);
		fprintf(hf, ":INPUT_ERROR:E\n");
		fflush(hf);
	}
	else {
		/*
		**  Allocate buffer for "user_input"
		*/

		buffer_size = sizeof(struct user_input);

		if ((user_query = (struct user_input *) malloc(buffer_size)) == NULL) {
			malloc_error("define.c", "term_input", "user_query", buffer_size);
		}

		/*
		**  ccopy "query_terms" to "source_terms" field
		*/
		
		sprintf(user_query->source_terms, "%s", query_terms);
		
		if (debugging) {
			fprintf(stderr, "term_input() :: define [%s]\n", query_terms);
		}
		
		/*
		**  Parse the user input with function 'parse_user_input'
		**  as follows:
		**
		**  1) Checks for adjacency operator (+) and removes it if present.
		**  2) Removes preceding and trailing spaces.
		**  3) Inserts a hyphen between first and second words if the
		**     first word is less than 4 chars long so that, if the first
		**     word is a stop-word it is not lost.
		*/
		
		parse_user_input(user_query);
		
		/*
		**  'superparse' user_query->source_terms
		*/
		
		sprintf(bss_command, "sp t=%s", user_query->source_terms);

		if (debugging) {
			fprintf(stderr, "term_input() :: bss_command = [%s]\n", bss_command);
		}
		
		if ((return_code = i0(bss_command, bss_result)) != 0) {
			terminate_run("term_input()", bss_command, return_code);
		}
		
		result_buffer = bss_result;
		
		if (debugging) {
			fprintf(stderr, "term_input() : bss_result: [%s]\n", result_buffer);
		}
		
		sscanf(result_buffer, "%d %n", &no_terms, &chars_read);
		result_buffer += chars_read;
		
		/*
		**  Store the number of terms in 'user_query->no_terms'
		*/
		
		user_query->no_terms = no_terms;
		
		/*
		**  Now extract component terms (parsed, gsl_class & source)
		*/
		
		for (current_term = 0; current_term < no_terms; current_term++) {
			result_buffer += 2;
			length = strstr(result_buffer, " c=") - result_buffer;
			memcpy(user_query->parsed[current_term], result_buffer, length);
			user_query->parsed[current_term][length] = '\0';
			
			result_buffer += length + 1;
			sscanf(result_buffer,
						 "c=%c s=%n",
						 &user_query->gsl_class[current_term], &chars_read);
			
			result_buffer += chars_read;
			
			if (current_term < no_terms - 1) {
				length = strstr(result_buffer, " t=") - result_buffer;
				memcpy(user_query->source[current_term], result_buffer, length);
				user_query->source[current_term][length] = '\0';
				result_buffer += length + 1;
			}
			else {
				length = strlen(result_buffer) - 1;
				sprintf(user_query->source[current_term], "%s", result_buffer);
				user_query->source[current_term][length] = '\0';
			}

			if (debugging) {
				fprintf(stderr, "term_input() :: TERM: [%s]\n",
								user_query->source[current_term]);
			}
		}
		
		if ((user_query->operation == NO_OP) || (no_terms == 1)) {
			/*
			**  This accounts for:
			**
			**  1) 1 or more terms, no adj operator:        operation = NO_OP
			**  2) A user defined GSL phrase:               operation = OP_GSL
			**  3) One term with a redundant adj operator:  operation = NO_OP
			**
			**  Set phrase fields to NONE_ASSIGNED
			*/

			user_query->parsed_phrase[0] = '\0';
			user_query->phrase_set = NONE_ASSIGNED;
			user_query->phrase_np = NONE_ASSIGNED;
			
			/*
			**  user_query->phrase_r = NONE_ASSIGNED;
			*/
			
			user_query->phrase_r = 0;
			user_query->phrase_wgt = (double) NONE_ASSIGNED;
			user_query->phrase_rsv = (double) NONE_ASSIGNED;
			
			for (current_term = 0; current_term < no_terms; current_term++) {
				/*
				**  First check for stop terms
				*/

				if (strchr(STOP_TERMS, user_query->gsl_class[current_term]) == NULL) {
					/*
					**  First check to see if term already exists
					*/
					
					existing_term_no =
						check_for_duplicates(user_query->parsed[current_term],
																 USER_TERM, &changed);
					
					if (existing_term_no == NONE_ASSIGNED) {
						/*
						**  It's a new term  set 'user_query->operation'
						**  to NO_OP or OP_GSL if necessary.
						*/
						
						if (strchr(user_query->source[current_term], ' ') == NULL) {
							user_query->operation = NO_OP;
						}
						else {
							user_query->operation = OP_GSL;
						}
						
						/*
						**  Write next individual term to history file.
						*/
						
						hf_header(COMMAND_NO, "define", PARAMS);
						
						if (user_query->operation == NO_OP) {
							fprintf(hf, ":%s:N\n", user_query->source[current_term]);
						}
						else {
							fprintf(hf, ":%s:G\n", user_query->source[current_term]);
						}
						
						fflush(hf);
						
						bss_find(user_query->parsed[current_term], &set, &npostings);
						
						user_query->term_np[current_term] = npostings;
						
						if (npostings > 0) {
							/*
							**  The term exists
							*/
							
							NEW_ITERATION = TRUE;
							
							user_query->term_set[current_term] = set;
							
							/*
							**  Determine little_r, weight and RSV.
							*/
							
							if (BIGR > 0) {
								user_query->term_r[current_term] =
									calc_little_r(user_query->term_set[current_term]);
							}
							else {
								user_query->term_r[current_term] = 0;
							}
							
							user_query->term_wgt[current_term] =
								calc_wgt(user_query->term_np[current_term],
												 user_query->term_r[current_term],
												 USER_TERM);
							
							user_query->term_rsv[current_term] =
								calc_rsv(user_query->term_wgt[current_term],
												 user_query->term_r[current_term],
												 USER_TERM);
							
							if (debugging) {
								fprintf(stderr, "term_input():: ");
								fprintf(stderr, "uq->wgt = [%.3f], uq->rsv = [%.3f]\n",
												user_query->term_wgt[current_term],
												user_query->term_rsv[current_term]);
							}
							
							/*
							**  Transfer user_term fields to termset
							*/
							
							(termset + TERM_NO)->term_no = TERM_NO;
							(termset + TERM_NO)->bss_set = user_query->term_set[current_term];
							(termset + TERM_NO)->np = user_query->term_np[current_term];
							(termset + TERM_NO)->little_r = user_query->term_r[current_term];
							(termset + TERM_NO)->weight = user_query->term_wgt[current_term];
							(termset + TERM_NO)->rsv = user_query->term_rsv[current_term];
							(termset + TERM_NO)->term_type = USER_TERM;
							(termset + TERM_NO)->attribute =
								tolower(user_query->gsl_class[current_term]);
							sprintf((termset + TERM_NO)->source, "%s",
											user_query->source[current_term]);
							sprintf((termset + TERM_NO)->parsed, "%s",
											user_query->parsed[current_term]);
							(termset + TERM_NO)->operation = user_query->operation;
							
							check_against_rels(TERM_NO);
							TERM_NO++;
						}
						else {
							/*
							**  np = 0, term does not exists in the database.
							**
							**  First convert operation value to lowercase and make
							**  (weight, rsv, bss_set) =
							**  (NONE_ASSIGNED, NONE_ASSIGNED, NONE_ASSIGNED)
							*/
							
							user_query->operation = tolower(user_query->operation);
							user_query->term_wgt[current_term] = (double) NONE_ASSIGNED;;
							user_query->term_rsv[current_term] = (double) NONE_ASSIGNED;;
							user_query->term_set[current_term] = NONE_ASSIGNED;
							
							/*
							**  Delete the empty set
							*/
							
							sprintf(bss_command, "delete %d", set);
							
							if ((return_code = i0(bss_command, bss_result)) != 0) {
								terminate_run("term_input()", bss_command, return_code);
							}
							
							/*
							**  Add term to 'not_found_list'
							**
							**  First add linefeed to 'not_found_list' if it is not empty
							*/

							if (strlen(not_found_list) > 0) {
								sprintf(not_found_list + strlen(not_found_list), "\n");

							}

							sprintf(not_found_list + strlen(not_found_list),
											"\"%s\" was not found.",
											user_query->source[current_term]);
						}
					}
					else {
						/*
						**  Already in termset - re-calculate r, weight & RSV.
						*/
						
						(termset + existing_term_no)->weight =
							calc_wgt((termset + existing_term_no)->np,
											 (termset + existing_term_no)->little_r,
											 (termset + existing_term_no)->term_type);
						
						(termset + existing_term_no)->rsv =
							calc_rsv((termset + existing_term_no)->weight,
											 (termset + existing_term_no)->little_r,
											 (termset + existing_term_no)->term_type);
						
						/*
						**  Add term to 'not_found_list' as:
						**
						**    "termA" exists as "termB"
						**
						**  if the sources are different.
						**
						**  First add linefeed to 'not_found_list' if it is not empty
						*/
						
						if (strcasecmp(user_query->source[current_term],
													 (termset + existing_term_no)->source) != 0) {

							if (strlen(not_found_list) > 0) {
								sprintf(not_found_list + strlen(not_found_list), "\n");
							}
							
							sprintf(not_found_list + strlen(not_found_list),
											"\"%s\" exists as \"%s\"",
											user_query->source[current_term],
											(termset + existing_term_no)->source);
						}
						else if (!changed) {
							/*
							**  changed will be TRUE if it existed as an extracted term.
							**  Don't add to not_found_list if this is the case.
							*/
							if (strlen(not_found_list) > 0) {
								sprintf(not_found_list + strlen(not_found_list), "\n");
							}

							sprintf(not_found_list + strlen(not_found_list),
											"\"%s\" is already in the query",
											user_query->source[current_term]);
						}
					}
				}
			}
		}
		else {
			/*
			**  A user-defined phrase that is not in the gsl.
			**
			**  form_phrase() is defined in utilities.c
			*/
			
			if (debugging) {
				fprintf(stderr, "term_input() :: USER-DEFINED PHRASE\n");
			}
			
			form_phrase(user_query, FALSE);
			
			/*
			**  Transfer appropriate fields of 'user_query' to 'termset'
			*/
			
			if (user_query->phrase_np > 0) {
				/*
				**  Phrase exists - termset modified.
				*/
				NEW_ITERATION = TRUE;
				
				(termset + TERM_NO)->term_no = TERM_NO;
				(termset + TERM_NO)->bss_set = user_query->phrase_set;
				(termset + TERM_NO)->np = user_query->phrase_np;
				(termset + TERM_NO)->little_r = user_query->phrase_r;
				(termset + TERM_NO)->weight = user_query->phrase_wgt;
				(termset + TERM_NO)->rsv = user_query->phrase_rsv;
				(termset + TERM_NO)->term_type = USER_TERM;
				(termset + TERM_NO)->attribute = tolower(USER_PHRASE);
				sprintf((termset + TERM_NO)->parsed, "%s", user_query->parsed_phrase);
				sprintf((termset + TERM_NO)->source, "%s", user_query->source_terms);
				(termset + TERM_NO)->operation = user_query->operation;
				
				if (debugging) {
					fprintf(stderr, "term_input() :: %d [%s] (%d)\n",
									(termset + TERM_NO)->bss_set,
									(termset + TERM_NO)->source,
									(termset + TERM_NO)->np);
				}

				/*
				**  Now check to see if term is in any of the rels.
				**  Set (termset + n)->relj[] to:
				**
				**    NONE_ASSIGNED if the relj is no longer in the database
				**    0 if not in relj
				**    1 if in relj
				*/
				
				check_against_rels(TERM_NO);
				TERM_NO++;
			}
			else {
				sprintf(not_found_list,
								"\"%s\" was not found.",
								user_query->source_terms);
			}
		}
		
		write_termset_file(termset, TERM_NO);
		free(user_query);
	}

  show_time("term_input", FINISH);
}

void parse_user_input(struct user_input *query) {
  /*
  **  Remove any spaces from the front and end of the user's query
  **  and to extract an adjacency operator if there is one.
  */

  int length;
  int pos;
  char *qbuf;
  char elapsed[RUNTIME_LENGTH];

  /*
  **  First look for adjacency operator
  */

  qbuf = query->source_terms;

  if (strchr(qbuf, '+') == NULL) {
    query->operation = NO_OP;
  }
  else {
    length = strchr(qbuf, '+') - qbuf;
    memcpy(query->source_terms, qbuf, length);
    query->source_terms[length] = '\0';
    query->operation = OP_ADJ;
  }
	
  de_space(query->source_terms);
	
  if (query->operation == OP_ADJ) {
    /*
    **  Test for length of first word.
    **  If < HYPHEN_POS characters (defined in .okapi_rc) long insert
		**  a hyphen between the first and second words to ensure that
    **  the first word is not thrown away if it is a stopword.
    */
		
		hf_header(COMMAND_NO, "define", PARAMS);
		fprintf(hf, ":%s:%c\n", query->source_terms, query->operation);
		fflush(hf);

    for (pos = 0; pos < strlen(query->source_terms); pos++) {
      if ((query->source_terms[pos] == ' ') && 
					(pos < atoi(getenv("HYPHEN_POS")))) {
				query->source_terms[pos] = '-';
      }
    }
  }
	
  /*
	**  Update elased time
	*/

  current_time(elapsed);
}


int check_for_duplicates(char parsed[], char user_type, int *changed) {
  /*
  **  Check to see if parsed already in list.
  **  Make sure operation field set to uppercase if it's there and np > 0
  **
  **  Set there to NONE_ASSIGNED by default.
  **  If it's there set there to the term_no in (termset)
  */
  
  int current_term;
  int existing_term_no = NONE_ASSIGNED;
  
  for (current_term = 0; current_term < TERM_NO; current_term++) {
    if (strcmp(parsed, (termset + current_term)->parsed) == 0) {
      if ((termset + current_term)->np > 0) {
				/*
				**  Reset opcode field to uppercase
				*/
				(termset + current_term)->operation =
					toupper((termset + current_term)->operation);
				/*
				**  Only need to re-calculate little_r for phrases. If it's a
				**  single term it will be correct anyway (incremented if
				**  necessary in the extract_terms() function.
				*/
				
				(termset + current_term)->weight =
					calc_wgt((termset + current_term)->np,
									 (termset + current_term)->little_r,
									 (termset + current_term)->term_type);
				(termset + current_term)->rsv =
					calc_rsv((termset + current_term)->weight,
									 (termset + current_term)->little_r,
									 (termset + current_term)->term_type);

				/*
				**  Set boolean variable *changed to true if it
				**  existed as an extractec terms
				*/

				*changed = ((termset + current_term)->term_type == EXTRACT_TERM);

				(termset + current_term)->term_type = user_type;
				
				existing_term_no = current_term;
				
				break;
      }
    }
  }
  return existing_term_no;
}

