/*
**  Trec 6 Okapi Interface
**
**  defs.h
**
**  mg - Mar 96, Sep 96, Oct 96, Mar 97

**  NOTE:
**
**  The macro 'G_EXTERN' is defined as nothing, i.e.
**
**  #define G_EXTERN
**
**  in gui.cc (or first program in any project) and as extern, i.e.
**
**  #define G_EXTERN extern
**
**  in all other progs.
*/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <malloc.h>
#include <signal.h>

#include <stdlib.h>
#include <defines.h>
#include <math.h>
#include <sys/time.h>

/*
**  Libraries for reading directory information
**  with readdir() and examining it with stat()
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

/*
**  tcl/tk libraries
*/

#include <tcl.h>
#include <tk.h>

#include <errno.h>

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#ifndef NULL
#define NULL 0
#endif

#ifndef SEEK_SET
#define SEEK_SET 0
#endif

/*
**  Parameters for array sizes for various objects
*/

#define TCL_BUFFER_SIZE (4 * 1024)
#define V_SMALL_BUFFER 256
#define SMALL_BUFFER 1024
#define OS_COMMAND_BUFFER_SIZE 1024
#define TITLE_BUFFER_SIZE 512

#define MAX_FIELDS 32
#define MAX_INDEXES 32
#define MAX_PASSAGES 2
#define MAX_HIGHLIGHT_TERMS 16384

#define REASONABLE_TERM_LENGTH 256
#define ERROR_MESSAGE_SIZE 256
#define RUNTIME_LENGTH 64

#define MAX_QUERY_SIZE 50
#define MIN_QUERY_SIZE 20

#define MAX_TERM_LENGTH 256

#define MAX_NUMBER_TERMS (2 * 1024)

/*
**  macros to flag different length docs/passages
*/

#define DOC_TOO_LONG 0
#define PASS_TOO_LONG 1
#define LENGTH_OK 2

#define EXTRA_CHARS 256

#define MAX_TERMS 256
#define MAX_DOCID_LENGTH 32
#define DATE_LENGTH 10
#define MAX_DATE_LENGTH 32
#define MAX_PHRASE_LENGTH 256

#define MAX_TITLE_LINES 3
#define COMMAND_SIZE 16
#define FILENAME_SIZE 256
#define MAX_DBNAME_LENGTH 16

#define MAX_NO_TOPICS 256

/*
**  initialisation file
*/

#define OKAPI_RC ".okapi_rc"

#define DATABASE_NAME_LENGTH 32

#define NONE_ASSIGNED -1

#undef MAXSETS
#define MAXSETS (16*1024)

/*
**  Directories and filenames
**
**  1. Directory in which to store profiles
*/

#define HISTORY_FILE "history"
#define TERMSET_FILE "termset"
#define RELS_FILE "relsfile"
#define ALL_ITERATIONS_FILE "all_iterations"
#define CURRENT_ITERATION_FILE "current_iteration"
#define WORKING_QUERY_FILE "working_query"
#define QUERY_DISPLAY_FILE "query_display"
#define EXTRACT_TERMS_FILE "extract_terms"
#define REMOVED_TERMS_FILE "removed_terms"
#define SHOW_FILE "extract_show"
#define HITLIST_FILE "hitlist"
#define FULL_RECS_FILE "full"

#define HITLIST_TITLES_FILE "hitlist_titles"

#define WRITE 0
#define READ 1
#define APPEND 2

/*
**  Flags for sorting the termset[]
*/

#define TSORT_RSV 0
#define TSORT_PSD 1

/*
**  Adjacency values:  A = ADJ, S = SAMES, N = No operation
**
**  These will be converted to lowercase if:
**
**  (a) np=0
**  (b) little_r becomes -ve
**  (c) the term is removed from the termset
**
**  'adj' operators:
**
**   adj  adjacent in forward order in same sentence.
**        possibly with intervening stopwords
**  adj2  adjacent in forward order in same sentence,
**        no intervening stopwords 
*/

#undef BSS_ADJ
#define BSS_ADJ "adj"

#undef BSS_ADJ2
#define BSS_ADJ2 "adj2"

#undef BSS_SAMES
#define BSS_SAMES "sames"

/*
**  domain values for operation field
*/

#undef OP_ADJ
#define OP_ADJ 'A'

#undef OP_SAMES
#define OP_SAMES 'S'

#undef OP_BOTH
#define OP_BOTH 'B'

#undef OP_GSL
#define OP_GSL 'G'

#undef NO_PHRASE
#define NO_PHRASE 'e'

#undef NO_OP
#define NO_OP 'N'

#define PHRASE_OPS "ASBG"
#define SHOW_OPS "ABSGN"

#define LEGAL_GSL "HFISGPN"

/*
**  STOP_TERMS changed from "FHI" to "FH" 11th April 1996 so that a
**  phrase like 'human rights' wont be rejected ('rights' parses to
**  'right' which has a gsl class of 'I'
**
**  Now only reject terms with gsl class codes of:
**
**    H      stop term
**    F      other function word (stopped) 
*/

#define STOP_TERMS "FH"

/*
**  Flags for term type: (user phrase, user term, extract term)
*/

#define USER_PHRASE 'P'
#define USER_TERM 'U'
#define EXTRACT_TERM 'E'

/*
**  cases for the occurrence of adj & sames sets
**
**  1) n(adj) = 0, n(sames) > 0
**  2) n(adj) > 0, n(sames) > 0, n(adj) == n(sames)
**  3) 0 < n(adj) < n(sames)
**  4) n(adj) = n(sames) = 0
*/

#define PHRASE_TYPE1 1
#define PHRASE_TYPE2 2
#define PHRASE_TYPE3 3
#define PHRASE_TYPE4 4

/*
**  Maximum displayed line length for both hitlist and fullrecs
*/

#define TITLE_CHARS 180

/*
**  Number of records displayed in a hitlist window
*/

#define MAX_HIT_TITLES 512

/*
**  Relevance judgements
*/

#define USER_REL 0

#define R_NONE 0
#define R_FULL 1
#define R_PASSAGE 2
#define R_NOT 3

#define WGT_SORT 0
#define RELNO_SORT 1

#ifdef START
#undef START
#endif

#ifdef FINISH
#undef FINISH
#endif

#define START 0
#define FINISH 1

/*
**  flags for writing command to history file:
**
**    NO_PARAMS - no extra parameters after the command
**    PARAMS    - 1 or more parameters after the command
*/

#define NO_PARAMS 0
#define PARAMS 1

/*
**  Unix commands for use in system() commands
*/

#define MV "mv -f"
#define CP "cp"
#define RM "rm -f"
#define SORT "sort"

/*
**  GUI Window geometry
*/

/**
#define FULL_WINDOW_X 1000
#define FULL_WINDOW_Y 675
#define FULL_WINDOW_OFFSET_X 0
#define FULL_WINDOW_OFFSET_Y 0
**/

/*
**  Global variables
*/

G_EXTERN int FEEDBACK;

/*
**
**  Array to hold the current database name.
**  Boolean variable to indicate whether a database is open or not.
*/

G_EXTERN char DATABASE_NAME[DATABASE_NAME_LENGTH];
G_EXTERN int DB_OPEN;
G_EXTERN int CLEAR_RF;

/*
**  File pointers
*/

G_EXTERN FILE *hf;  /* history_file                                     */
G_EXTERN FILE *tf;  /* termsets_file                                    */
G_EXTERN FILE *rf;  /* relsfile                                         */
G_EXTERN FILE *sf;  /* show_file                                        */
G_EXTERN FILE *ef;  /* extract_terms file                               */
G_EXTERN FILE *qdf; /* query display file to be loaded into text widget */
G_EXTERN FILE *wqf; /* working query file                               */
G_EXTERN FILE *cif; /* current iteration file                           */
G_EXTERN FILE *aif; /* all iterations file                              */

/*
**  Arrays for filenames
**
**  First the profile files
*/

G_EXTERN char LOGS_DIR[FILENAME_SIZE];
G_EXTERN char history_file[FILENAME_SIZE];
G_EXTERN char termset_file[FILENAME_SIZE];
G_EXTERN char rels_file[FILENAME_SIZE];

/*
**  Now the working files
*/

G_EXTERN char show_file[FILENAME_SIZE];
G_EXTERN char hitlist_file[FILENAME_SIZE];
G_EXTERN char fullrecs_file[FILENAME_SIZE];
G_EXTERN char working_query_file[FILENAME_SIZE];
G_EXTERN char current_iteration_file[FILENAME_SIZE];
G_EXTERN char all_iterations_file[FILENAME_SIZE];

/*
**  Command buffers
*/

G_EXTERN char OS_COMMAND[OS_COMMAND_BUFFER_SIZE];
G_EXTERN char ERROR_MESSAGE[MAX_PHRASE_LENGTH];
G_EXTERN char LAST_COMMAND[COMMAND_SIZE];

G_EXTERN char TOPIC_NAME[FILENAME_SIZE];

/*
**  Incremental query expansion parameters
**  Initialised in utilities.c:initialise_global_variables().
**
**  their values may be altered during the program run
*/

G_EXTERN int LR_THRESHOLD;
G_EXTERN double RSV_FACTOR;

/*
**  Variables for displaying counts of relevance judgements
*/

G_EXTERN int QUERY_SIZE;
G_EXTERN int RJ_COUNT;

G_EXTERN int FIRST_REC;
G_EXTERN int LAST_REC;

G_EXTERN int FINAL_SEARCH;

/*
**  NEW_ITERATION is a boolean variable to indicate whether new
**  terms have been added to the termset since the last search.
**
**   TRUE: new terms have been added.
**  FALSE: new terms have not been added.
**
**  It is intialised to TRUE in initialise.c
**  It is set to FALSE in 'search.c'
**  It is set to TRUE in functions 'do_termset()' and 'do_extract()'
*/

G_EXTERN int NEW_ITERATION;

/*
**  Boolean variable to indicate whether a profile is
**  currently being edited.
*/

G_EXTERN int EDITING_TOPIC;

/*
**  EXTRACTED_TERMS is initialised to FALSE in initialisation(),
**  defined in 'initialise.c'.
**
**  It is set to TRUE every time 'extract_terms()' is entered
**  and to FALSE at the end of 'expand_query()'.
*/

G_EXTERN int EXTRACTED_TERMS;

/*
**  Time variables
*/

G_EXTERN long LAST_TIME;
G_EXTERN long ELAPSED_TIME;

/*
**  Variables to hold the current show set,
**  record number, weight and docid
*/

G_EXTERN int SHOW_SET;
G_EXTERN int SHOW_ABS_RECNO;
G_EXTERN int SHOW_SET_RECNO;
G_EXTERN double SHOW_WGT;
G_EXTERN char SHOW_DATE[DATE_LENGTH];
G_EXTERN char SHOW_DOCID[MAX_DOCID_LENGTH];

/*
**  NEWTERMS will be TRUE is new terms have been added to the
**  current termset but a new search has not yet been conducted.
**  It will be set to FALSE on the performance of a new search.
*/

G_EXTERN int NEW_TERMS;

/*
**  Variables to increment the temporary hitlist and show files
*/

G_EXTERN int HIT_COUNT;
G_EXTERN int SHOW_COUNT;

/*
**  The set number (always 0) and rel count for the
**  set of relevance judgements
*/

G_EXTERN int BIGR_SET;
G_EXTERN int BIGR;
G_EXTERN int SEEN_SET;
G_EXTERN int NO_SEEN;

/*
**  The topic no of the current search
*/

G_EXTERN int TOPIC_NO;

/*
**  Counts of terms entered (by the user or by an extract)
**  into the termset file.
**
**  TERM_COUNT is used so that the user can obtain up-to-date
**  weighting information about the query terms used to generate
**  the current document set after 1 or more rels have been made.
**
**  TERNSET_NO is used only when writing terms to the history file.
**  It is incremented in do_display().
*/

G_EXTERN int TERM_NO;
G_EXTERN int TERM_COUNT;
G_EXTERN int ITERATION_NO;
G_EXTERN int NO_REMOVED_TERMS;
G_EXTERN int RELJ_COUNT;

/*
**  The current command_no for writing to the history file.
**  Used as the primary key in the Oracle tables.
*/

G_EXTERN int COMMAND_NO;

/*
**  TOPIC_COUNT is the index, in TOPIC_LIST, of the current profile
*/

G_EXTERN int TOPIC_COUNT;
G_EXTERN int NO_OF_TOPICS;

/*
**  Character array to hold profile names.
**  TOPIC_LIST[0] = default name i.e. <user_id>.<pid>
**  user named profiles will go in index positions
**  1 .. MAX_NO_TOPICS
*/

G_EXTERN char TOPIC_LIST[MAX_NO_TOPICS + 1][FILENAME_SIZE];


/*
**  Structure to hold parameter file data
*/

G_EXTERN struct params_file {
	char name[V_SMALL_BUFFER];
	int lastbibvol;
	char bib_basename[V_SMALL_BUFFER];
	char bib_dir[V_SMALL_BUFFER];
	int bibsize;
	int real_bibsize;
	char display_name[V_SMALL_BUFFER];
	char explanation[V_SMALL_BUFFER];
	int nr;
	int nf;
	char f_abbrev[V_SMALL_BUFFER][MAX_FIELDS];
	int rec_mult;
	int fixed;
	char db_type[V_SMALL_BUFFER];
	int has_lims;
	int maxreclen;
	int ni;
	int last_ixvol[MAX_INDEXES];
	char ix_stem[V_SMALL_BUFFER][MAX_INDEXES];
	int ix_volsize[MAX_INDEXES];
	int ix_type[MAX_INDEXES];
	int para_file;
} params_file;

/*
**  Structure to hold all information about members of the 
**  cumulative termset.
*/

typedef struct query_terms {
  int term_no;         /*  RSV order sequence number (for use in
											 **	 the remove function
											 */
  int bss_set;         /*  the document set generated by this termset */
  int np;              /*  No. postings                               */
  int little_r;        /*  number of documents containing the term    */
  double weight;       /*  F4 weight                                  */
  double rsv;          /*  Robertson selection value                  */
  char term_type;      /*  [USER_TERM, EXTRACT_TERM]                  */
  char attribute;      /*  [<gsl_class>, USER_PHRASE]                 */
  char parsed[MAX_TERM_LENGTH];  /*  parsed term                 */
  char source[MAX_TERM_LENGTH];  /*  source of parsed term       */
  char operation;      /*  [OP_ADJ, OP_SAMES, OP_BOTH, OP_GSL, NO_OP] */
  int relj[V_SMALL_BUFFER];
                       /*  0..BIGR, each 0 or 1 indicating in rel[]   */
} query_terms;

G_EXTERN struct query_terms *termset;

/*
**  Working query
*/

typedef struct working_query {
  char term_type;
  int bss_set;
  int np;
  double weight;
  int little_r;
  double rsv;
  char source[MAX_TERM_LENGTH];
  char parsed[MAX_TERM_LENGTH];
  char operation;
  char start_tag;
  char end_tag;
} working_query;


/*
**  Structure to hold the <= 20 sets to be used in the next search.
*/

typedef struct search_terms {
  int bss_set;
  double weight;
  char operation;
} search_terms;

/*
**  Structure declaration for holding user input. This will hold the
**  full user input -- in terms[] -- the source and parsed forms of
**  each component term together with each gsl class -- in source[][],
**  parsed[][] and gsl_class.
**
**  For a user-defined phrase the bss set, np, r, wgt and rsv will be
**  held in bss_set, np, r, wgt & rsv fields
*/

typedef struct user_input {
  char source_terms[MAX_TERM_LENGTH];
  char parsed_phrase[MAX_TERM_LENGTH];
  char operation;
  int no_terms;
  int phrase_set;
  int phrase_np;
  int phrase_r;
  double phrase_wgt;
  double phrase_rsv;
  char source[MAX_TERMS][MAX_TERM_LENGTH];
  char parsed[MAX_TERMS][MAX_TERM_LENGTH];
  char gsl_class[MAX_TERMS];
  int term_set[MAX_TERMS];
  int term_np[MAX_TERMS];
  int term_r[MAX_TERMS];
  double term_wgt[MAX_TERMS];
  double term_rsv[MAX_TERMS];
} user_input;


/*
**  Shown documents structure: includes F, P & N rels.
**
**  set = bm set,
**  record = ranked position within set
**  rj = R_FULL (1), R_PASSAGE (2), R_NOT (3)
**  weight = docs weight within the set
**
**  date is recorded for future use - i.e. to remove or
**  reduce the effect of information from a relevance
**  judgement that is no longer in the database.
*/

typedef struct user_rj {
  int relno;                     /*  sequence number of original relj    */
  int set;                       /*  bss_set of hitlist doc set          */
  int abs_recno;                 /*  BSS internal record number          */
  int set_recno;                 /*  sequence_no within SEEN_SET         */
  int rj;                        /*  relevance judgement                 */
  double weight;                 /*  doc weight when last +ve rel made   */
  int offset;                    /*  offset of relevant part of document */
  int length;                    /*  length of relevant part of document */
	char sflag;                    /*  'C' - part of current search
																 **  'N' - a rel made before clearing
																 **        the query or just rf 
																 */
  char docid[MAX_DOCID_LENGTH];  /*  docid                               */
  char title[MAX_PHRASE_LENGTH]; /*  title displayed in hitlist          */
  char date[MAX_DATE_LENGTH];    /*  date field of document              */
} user_rj;

G_EXTERN struct user_rj *user_rels;

G_EXTERN int USER_RELS_COUNT;


/*
**  Structure to hold the hitlist titles generated by a document set
*/

G_EXTERN struct hit_titles {
  int abs_recno;
  int set_recno;
  char docid[MAX_DOCID_LENGTH];
  char title[TITLE_CHARS];
} hit_titles[MAX_HIT_TITLES];


/*
**  Structure to hold the hitlist info for a given
**  entry in the hitlist.
*/

typedef struct hitlist_info {
  char docid[MAX_DOCID_LENGTH];
  int set_recno;
	int abs_recno;
  double weight;
  int first_row;
  int last_row;
  int offset[2];  /* [0] = passage */
  int length[2];  /* [1] = fullrec */
} hitlist_info;

/*
**  NTD's defs for tk programs.
**
**  Name of a user-specific startup script
*/

G_EXTERN char tcl_RcFile[FILENAME_SIZE];

#ifndef ENQ_DEC_DOT_H
#define ENQ_DEC_DOT_H

/*
**  XXX
*/

EXTERN int OKAPI_Init();
EXTERN int OKAPI_Exit();

/*
**  QUERY
*/

/*
**  EXTERN int QM_Init();
**  EXTERN int QM_Exit();
*/

/*
**  ENQ_UTILS
*/

EXTERN int UTILS_Init();
EXTERN int UTILS_Exit();

/*
**  ENQUIRE
*/

EXTERN int Do_AppDeConstruct();

#endif

/*
**  End of NTD's defs
*/


/*
**  Structures to hold new show format header information
*/

typedef struct header {
  int msg_len;           /* header + document                            */
  char reserved[32];
  int setnum;            /* bss set                                      */
  int set_recno;         /* rank within set                              */
  int abs_recno;         /* Absolute record number within database       */
  double weight;
  char unused[10];
  int data_offset;       /* offset from start of header to start of data */
  int no_dir_recs;       /* title, hl, etc.                              */
  int no_passage_recs;   /* 0 or 2                                       */
  int no_highlight_recs; /*  number of highlighted terms                 */
} header;

/*
**  The remaining 3 structures will hold the directory,
**  passage and highlight records
*/

typedef struct dir_records {
  int field_no;
  int offset;
  int length;
} dir_records;

typedef struct passage_records {
  double weight;
  int offset;
  int length;
} passage_records;

typedef struct highlight_records {
  char code;
  int offset;
  int length;
} highlight_records;

/*
**  Structure to store current query terms and removed terms
*/

typedef struct query_list_lines {
  char term[REASONABLE_TERM_LENGTH];
} query_list_lines;


/*
**  Structure in which to read log file directory information.
*/

typedef struct logfile_info {
  char logfile[FILENAME_SIZE];
} logfile_info;

