/*
**  display.c
**
**  display_terms() function. Moved from utilities.c
**
**  mg - Apr 1997
*/

#define G_EXTERN extern

#include "defs.h"

/*
**  Length of query term before it needs to be wrapped
*/

#define DISPLAY_OPS "SB"

G_EXTERN void show_time(char *, int);
G_EXTERN void sort_terms(struct query_terms *, int, int);
G_EXTERN void sort_current_query(struct working_query *, int);

G_EXTERN int calc_little_r(int bss_set);
G_EXTERN double calc_wgt(int np, int r, char term_type);
G_EXTERN double calc_rsv(double weight, int r, char term_type);

/*
**  This function determines which terms constitute the current
**  query before sorting them into descending RSV order, and
**  writing them to the query_display_file which is to be loaded
**  into the appropriate interface widget.
*/

char *display_terms(struct query_list_lines query_list[],
										int *query_list_count) {
  /*
  **  Need to re-calculate little_r, weight and rsv for phrases
	**  since this wont be done when terms are extracted.
	**
  **  The termset structure will be sorted in this function.
  **
  **  11 Jun 1996 - the conditions for terms to be included in the 
  **  working query - i.e. must be:
  **
  **  a) a user term
  **  b) an extracted term that has:
	**
	**      little_r >= LR_THRESHOLD
  **      RSV >= RSV_FACTOR * AVG(current RSV of all terms that 
	**                              have been in the working query)
	**
  **     e.g. EXTRACT_TERMS_THRESHOLD = 2, and RSV_FACTOR = 0.67
	**  
	**  LR_THRESHOLD & RSV_FACTOR are global variables, initialised in
	**  utilities.c:initialise_global_variables().
  */

	int debugging = atoi(getenv("DISPLAY_DEBUG"));
	
	char term_type;
	
	int bss_set;
	int little_r;
	int np;
	
	int window_width = atoi(getenv("QUERY_WINDOW_WIDTH"));

	double rsv;
	double weight;
  double rsv_total = 0.0;
	double average_rsv;

  int iloop;
  int next_term = 0;
  int no_terms;
  int term_count = 0;
  int last_iteration_count = 0;
  int tloop;

  int in_previous_iteration;
  char start_tag;
  char end_tag;

  char temp[SMALL_BUFFER];
  char query_display_file[FILENAME_SIZE];

  struct working_query current_query[MAX_QUERY_SIZE];

  char *in;

  show_time("display_terms", START);

  /*
	**  Terms not found (np == 0) and terms that have been removed
  **  by the user will have the operation field converted to lowercase.
  **
  **  As well as the conditions defined above, only terms with an
  **  uppercase value in the operation field (defined in SHOW_OPS)
  **  will be displayed.
  **
  **  First sort all terms into the order:
  **
	**    user_type D, rsv D, weight D, little_r, np
  */
  
  if (TERM_NO > 1) {
    sort_terms(termset, TERM_NO, TSORT_RSV);
  }
	
	/*
	**  If first iteration (0) or after (ITERATION_NO > NONE_ASSIGNED)
  **  read in information about the last iteration (set, weight & np)
  */
	
	if ((ITERATION_NO > NONE_ASSIGNED) && (!CLEAR_RF)) {
		/*
		**  Open "all_iterations_file" for reading.
		**
		**  Read file to determine average RSV of all
		**  terms that have previously been in the query.
		*/
		
		if ((aif = fopen(all_iterations_file, "r")) == NULL) {
			printf("ERROR OPENING FILE [%s]\n", all_iterations_file);
			exit(0);
		}
		
		rsv_total = 0.0;
		
    while (!feof(aif)) {
      if (fgets(temp, SMALL_BUFFER, aif) == NULL) {
				break;
      }
      else {
				temp[strlen(temp) - 1] = '\0';
				in = temp;
				sscanf(in, "%d:%d:%c\n", &bss_set, &np, &term_type);
				
				little_r = calc_little_r(bss_set);
				weight = calc_wgt(np, little_r, term_type);
				rsv = calc_rsv(weight, little_r, term_type);
				term_count++;
				rsv_total += rsv;
      }
    }
		
    fclose(aif);
		
		average_rsv = RSV_FACTOR * (rsv_total / (double) term_count);
  }
	else {
		/*
		**  Set average_rsv to zero if there has not yet been
		**  a search (i.e. ITERATION_NO = NONE_ASSIGNED).
		*/
		average_rsv = 0.0;
		CLEAR_RF = FALSE;
	}

	if (debugging) {
		fprintf(stderr, "display_terms():: average_rsv = %f\n",
						average_rsv);
	}

  /*
  **  Now transfer valid terms to the current_query structure.
	**  Also append to the all_iterations_file the "bss_set",
	**  "np" and "term_type" for each valid term.
  */
	
	aif = fopen(all_iterations_file, "a");

  next_term = 0;
	term_count = 0;


	while (term_count < TERM_NO) {
    if (strchr(SHOW_OPS, (termset + term_count)->operation) != NULL) {
      /*
      **  i.e. it has an uppercase operation field.
      */

			if ((toupper((termset + term_count)->term_type) == USER_TERM) ||
					((toupper((termset + term_count)->term_type) == EXTRACT_TERM) &&
					 ((termset + term_count)->little_r >= LR_THRESHOLD) &&
					 ((termset + term_count)->rsv >= average_rsv))) {
				/*
				**  Term satisfies the conditions for membership of the query
				**  Re-calculate little_r, weight & rsv for phrases as
				**  this is not done by the extract_terms() function.
				*/
				
				if (strchr(PHRASE_OPS, (termset + term_count)->operation) != NULL) {
					(termset + term_count)->little_r =
						calc_little_r((termset + term_count)->bss_set);

					(termset + term_count)->weight =
						calc_wgt((termset + term_count)->np,
										 (termset + term_count)->little_r,
										 (termset + term_count)->term_type);
					
					(termset + term_count)->rsv =
						calc_rsv((termset + term_count)->weight,
										 (termset + term_count)->little_r,
										 (termset + term_count)->term_type);
				}

				current_query[next_term].term_type = (termset + term_count)->term_type;
				current_query[next_term].bss_set = (termset + term_count)->bss_set;
				current_query[next_term].np = (termset + term_count)->np;
				current_query[next_term].little_r = (termset + term_count)->little_r;
				current_query[next_term].weight = (termset + term_count)->weight;
				current_query[next_term].rsv = (termset + term_count)->rsv;
				current_query[next_term].operation = (termset + term_count)->operation;
				
				sprintf(current_query[next_term].source, "%s",
								(termset + term_count)->source);
				
				sprintf(current_query[next_term].parsed, "%s",
								(termset + term_count)->parsed);
				
				if (debugging) {
					fprintf(stderr,
									"display_terms():: current_query[%d].term_type = [%c]\n",
									next_term, current_query[next_term].term_type);
					fprintf(stderr,
									"display_terms():: current_query[%d].bss_set = [%d]\n",
									next_term, current_query[next_term].bss_set);
					fprintf(stderr,
									"display_terms():: current_query[%d].np = [%d]\n",
									next_term, current_query[next_term].np);
					fprintf(stderr,
									"display_terms():: current_query[%d].little_r = [%d]\n",
									next_term, current_query[next_term].little_r);
					fprintf(stderr,
									"display_terms():: current_query[%d].weight = [%.3f]\n",
									next_term, current_query[next_term].weight);
					fprintf(stderr,
									"display_terms():: current_query[%d].rsv = [%.3f]\n",
									next_term, current_query[next_term].rsv);
					fprintf(stderr,
									"display_terms():: current_query[%d].operation = [%c]\n",
									next_term, current_query[next_term].operation);
				}

				/*
				**  Append (bss_set, np, term_type) to all_iterations_file.
				**  Need these values to re-calculate (little_r, wgt, rsv)
				**  for each term to include in the average_rsv total next
				**  time around.
				*/

				fprintf(aif, "%d:%d:%c\n",
								current_query[next_term].bss_set,
								current_query[next_term].np,
								current_query[next_term].term_type);

				/*
				**  Use the term_type field as a flag as to whether
				**  this is a new (isupper) or an old term (islower)
				*/
				
				if (isupper((termset + term_count)->term_type)) {
					current_query[next_term].start_tag = 'S';
					current_query[next_term].end_tag = 'E';
					
					/*
					**  Set 'term_type' field to lowercase to indicate
					**  that, in future, this is an 'old' term.
					*/
					
					(termset + term_count)->term_type =
						tolower((termset + term_count)->term_type);
				}
				else {
					current_query[next_term].start_tag = 'T';
					current_query[next_term].end_tag = 'W';
				}
				next_term++;
      }
    }

    if (next_term == QUERY_SIZE) {
      /*
      **  current_query is full
      */
      break;
    }

    term_count++;
  }

	fclose(aif);

  no_terms = next_term;
	
  /*
  **  Write query_display_file, to be loaded into 'query_list' widget
  */

  if (no_terms > 0) {
    /*
    **  Sort current_query into desc rsv order.
    */
    
    if (no_terms > 1) {
      sort_current_query(current_query, no_terms);
		}
    
    /*
    **  Now write query_display_file which will be loaded into the
    **  query text widget, and fill out the query_list array which
    **  is returned to the calling function so that the query list
    **  is available to the remove_terms function. Note, if the
    **  query entry is too long to fit onto one line in the query
    **  text widget, we will write it to 2 consecutive locations
    **  in query_list so that clicking on either line will enable
    **  the correct term to be located.
    */

    sprintf(query_display_file, "%s/%s.%d",
						Temp_dir, QUERY_DISPLAY_FILE, getpid());
    
    if ((qdf = fopen(query_display_file, "w")) == NULL) {
      printf("ERROR opening file [%s]\n", query_display_file);
      exit(0);
    }
		
    if ((wqf = fopen(working_query_file, "w")) == NULL) {
      printf("ERROR opening file [%s]\n", working_query_file);
      exit(0);
    }
		
    *query_list_count = 0;
		
    for (tloop = 0; tloop < no_terms; tloop++) {
			fprintf(qdf, "%7d%3d : <%c>%s",
							current_query[tloop].np,
							current_query[tloop].little_r,
							current_query[tloop].start_tag,
							current_query[tloop].source);
			
      sprintf((query_list + (*query_list_count)++) -> term,
							"%s", current_query[tloop].source);
			
			/*
      if (strlen(current_query[tloop].source) > window_width) {
				sprintf((query_list + (*query_list_count)++) -> term,
								"%s", current_query[tloop].source);
      }
			*/
      
			/*
			**  Only show opcode if in [BS], i.e. not if in [AG]
			*/

			if (strchr(DISPLAY_OPS, current_query[tloop].operation) != NULL) {
				fprintf(qdf, " (%c)", current_query[tloop].operation);
      }

			fprintf(qdf, "<%c>\n", current_query[tloop].end_tag);
			
      fprintf(wqf, "%c:%d:%d:%.3f:%c:%s:%s\n",
							current_query[tloop].term_type,
							current_query[tloop].bss_set,
							current_query[tloop].np,
							current_query[tloop].weight,
							current_query[tloop].operation,
							current_query[tloop].source,
							current_query[tloop].parsed);

			if (debugging) {
				fprintf(stderr,
								"display_terms():: current_query[%d].term_type = [%c]\n",
								tloop, current_query[tloop].term_type);
				fprintf(stderr,
								"display_terms():: current_query[%d].bss_set = [%d]\n",
								tloop, current_query[tloop].bss_set);
				fprintf(stderr,
								"display_terms():: current_query[%d].np = [%d]\n",
								tloop, current_query[tloop].np);
				fprintf(stderr,
								"display_terms():: current_query[%d].weight = [%.3f]\n",
								tloop, current_query[tloop].weight);
				fprintf(stderr,
								"display_terms():: current_query[%d].operation = [%c]\n",
								tloop, current_query[tloop].operation);
				fprintf(stderr,
								"display_terms():: current_query[%d].source = [%s]\n",
								tloop, current_query[tloop].source);
				fprintf(stderr,
								"display_terms():: current_query[%d].parsed = [%s]\n",
								tloop, current_query[tloop].parsed);
			}
		}

    fclose(qdf);
    fclose(wqf);
  }
  else {
    /*
    **  Empty query
    */
    sprintf(query_display_file, "EMPTY");
  }
	
	show_time("display_terms", FINISH);

	return query_display_file;
}

