/*
**  gui.cc
**
**  mg
**
**   7th Jun 1995, adapted from ntd's gui.cc file
**  28th May 1996 - modified to use text widgets for the hitlist
**                  and rels pool windows instead of listboxes.
**
**   7th Jul 1996 - modified to use a text widget for:
**
**                  a) current query window.
**                  b) removed_terms window.
**
*/


#define _XXX_MODULE

#define G_EXTERN

#define OKAPI_MODULE

#include "defs.h"
#include "GLOB.h"

#define ENABLE_SEARCH 1
#define DISABLE_SEARCH 0

#define ENABLE_OPTIMISE 1
#define DISABLE_OPTIMISE 0

#define HITLIST_WINDOW 0
#define RELSPOOL_WINDOW 1

/*
**  'C' functions linked with TCL procedures
*/

extern int adjust_rsv_factor(ClientData clientData, Tcl_Interp *interp, 
														 int argc, char *argv[]);

extern int clear_query(ClientData clientData, Tcl_Interp *interp, 
											 int argc, char *argv[]);

extern int clear_rf(ClientData clientData, Tcl_Interp *interp, 
										int argc, char *argv[]);

extern int db_search(ClientData clientData, Tcl_Interp *interp, 
										 int argc, char *argv[]);

extern int make_rel(ClientData clientData, Tcl_Interp *interp, 
										int argc, char *argv[]);

extern int query_expand(ClientData clientData, Tcl_Interp *interp, 
												int argc, char *argv[]);

extern int quit_gui(ClientData clientData, Tcl_Interp *interp, 
										int argc, char *argv[]);

extern int reinstate_term(ClientData clientData, Tcl_Interp *interp, 
													int argc, char *argv[]);

extern int remove_term(ClientData clientData, Tcl_Interp *interp, 
											 int argc, char *argv[]);

extern int set_lr_threshold(ClientData clientData, Tcl_Interp *interp, 
														int argc, char *argv[]);

extern int set_query_size(ClientData clientData, Tcl_Interp *interp, 
													int argc, char *argv[]);

extern int set_rsv_factor(ClientData clientData, Tcl_Interp *interp, 
													int argc, char *argv[]);

extern int show_rec(ClientData clientData, Tcl_Interp *interp, 
										int argc, char *argv[]);

extern int show_rels(ClientData clientData, Tcl_Interp *interp, 
										 int argc, char *argv[]);

extern int term_entry(ClientData clientData, Tcl_Interp *interp, 
											int argc, char *argv[]);

extern int window_title(ClientData clientData, Tcl_Interp *interp, 
												int argc, char *argv[]);

/*
**  Global variables for this program
*/

static int query_list_count;

struct query_list_lines query_list[2*MAX_QUERY_SIZE];

static struct hitlist_info hitlist_index[2 * SMALL_BUFFER];
static struct hitlist_info relspool_index[2 * SMALL_BUFFER];

static int last_hitlist_row;
static int last_relspool_row;

static int hitlist_fcount;
static int relspool_fcount;

struct query_list_lines removed_terms[SMALL_BUFFER];

static int SHOW_INDEX;

static double maximum_weight;

static full_doclen;
static part_doclen;

int OKAPI_Init()
{
	
  /*
  **  tcl -> c  function mappings
  **
  **  c.<name> is a tcl pointer to the C function <name>
  **
  **  theInterp           tcl interpreter
  **  adjust_rsv_factor   c.adjust_rsv_factor
  **  clear_query         c.clear_query
  **  clear_rf            c.clear_rf
  **  db_search           c.db_search
  **  make_rel            c.make_rel
  **  query_expand        c.query_expand
  **  quit_gui            c.quit_gui
  **  reinstate_term      c.remove_term
  **  remove_term         c.remove_term
  **  set_lr_threshold    c.set_lr_threshold
  **  set_query_size      c.set_query_size
  **  set_rsv_factor      c.set_rsv_factor
  **  show_rec            c.show_rec
  **  show_rels           c.show_rels
  **  term_entry          c.term_entry
  **  window_title        c.window_title
  */
  
  Tcl_CreateCommand(theInterp, "c.adjust_rsv_factor", adjust_rsv_factor,
										(ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
  
  Tcl_CreateCommand(theInterp, "c.clear_query", clear_query,
										(ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
  
  Tcl_CreateCommand(theInterp, "c.clear_rf", clear_rf,
										(ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
  
  Tcl_CreateCommand(theInterp, "c.db_search", db_search,
										(ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
  
  Tcl_CreateCommand(theInterp, "c.make_rel", make_rel,
										(ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
	
  Tcl_CreateCommand(theInterp, "c.query_expand", query_expand,
										(ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
	
  Tcl_CreateCommand(theInterp, "c.quit_gui", quit_gui,
										(ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
	
  Tcl_CreateCommand(theInterp, "c.reinstate_term", reinstate_term,
										(ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
	
  Tcl_CreateCommand(theInterp, "c.remove_term", remove_term,
										(ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
	
  Tcl_CreateCommand(theInterp, "c.set_lr_threshold", set_lr_threshold,
										(ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
	
  Tcl_CreateCommand(theInterp, "c.set_query_size", set_query_size,
										(ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
	
  Tcl_CreateCommand(theInterp, "c.set_rsv_factor", set_rsv_factor,
										(ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
	
  Tcl_CreateCommand(theInterp, "c.show_rec", show_rec,
										(ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
	
  Tcl_CreateCommand(theInterp, "c.show_rels", show_rels,
										(ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
	
  Tcl_CreateCommand(theInterp, "c.term_entry", term_entry,
										(ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
	
  Tcl_CreateCommand(theInterp, "c.window_title", window_title,
										(ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
  
  hitlist_fcount = relspool_fcount = 0;
	
  return TCL_OK;
}


int OKAPI_Exit()
{

  /*
	**  clear up before exit
	*/
  
  Tcl_DeleteCommand(theInterp, "c.adjust_rsv_factor");
  Tcl_DeleteCommand(theInterp, "c.clear_query");
  Tcl_DeleteCommand(theInterp, "c.clear_rf");
  Tcl_DeleteCommand(theInterp, "c.db_search");
  Tcl_DeleteCommand(theInterp, "c.make_rel");
  Tcl_DeleteCommand(theInterp, "c.query_expand");
  Tcl_DeleteCommand(theInterp, "c.quit_gui");
  Tcl_DeleteCommand(theInterp, "c.reinstate_term");
  Tcl_DeleteCommand(theInterp, "c.remove_term");
  Tcl_DeleteCommand(theInterp, "c.set_lr_threshold");
  Tcl_DeleteCommand(theInterp, "c.set_query_size");
  Tcl_DeleteCommand(theInterp, "c.set_rsv_factor");
  Tcl_DeleteCommand(theInterp, "c.show_rec");
  Tcl_DeleteCommand(theInterp, "c.show_rels");
  Tcl_DeleteCommand(theInterp, "c.term_entry");
  Tcl_DeleteCommand(theInterp, "c.window_title");

  return TCL_OK;

}


/*
**  TCL Link functions
**
**
**  db_search()
**
**  Calls 'search.c :: bss_search()' to search the database
**  on the current query.
**  Calls the function 'build_hitlist' which returns a filename
**  which is passed back to the Tcl function 'db_search'
*/

void test_fcn(int show_set, int np, double maximum_weight, int window,
							char hitlist_file[]) {
	printf("%5d %5d %.3f %3d  [%s]\n",
				 show_set, np, maximum_weight, window, hitlist_file);
}


int adjust_rsv_factor(ClientData clientData, Tcl_Interp *interp, 
											int argc, char *argv[]) {

	int debug = atoi(getenv("ADJUST_RSV_FACTOR_DEBUG"));

	double old_rsv_factor;

  char direction;
	char error_message[64];

	char query_display_file[FILENAME_SIZE];
	char gui_result[SMALL_BUFFER];

  hf_header(COMMAND_NO++, "adjust_rsv_factor", PARAMS);

	sprintf(error_message, "NONE");

	if (debug) {
		fprintf(stderr, 
						"set_rsv_factor() :: old factor = %lf\n",
						RSV_FACTOR);
	}

	old_rsv_factor = RSV_FACTOR;

	if (argc == 2) {
		/*
		**  The new (changed) RSV_FACTOR is passed to this function
		*/
		direction = argv[1][0];

		switch (direction) {
  		case 'D' :
				if (RSV_FACTOR >= 0.1) {
					RSV_FACTOR -= 0.1;
				}
				else {
					sprintf(error_message,
									"Cannot reduce RSV_FACTOR from %.2f", RSV_FACTOR);
				}

				break;

  		case 'U' :
				if (RSV_FACTOR <= 0.9) {
					RSV_FACTOR += 0.1;
				}
				else {
					sprintf(error_message,
									"Cannot increase RSV_FACTOR from %.2f", RSV_FACTOR);
				}

				break;
		}

		if (strcmp(error_message, "NONE") == 0) {
			fprintf(hf, ":%lf\n", RSV_FACTOR);

			if (debug) {
				fprintf(stderr, "adjust_rsv_factor() :: %lf to %lf\n",
								old_rsv_factor, RSV_FACTOR);
			}
		}
		else {
			fprintf(hf, ":%s\n", error_message);

			if (debug) {
				fprintf(stderr, "adjust_rsv_factor() :: %s\n", error_message);
			}
		}

		if (old_rsv_factor != RSV_FACTOR) {
			if (TERM_NO > 0) {
				sprintf(query_display_file, "%s",
								display_terms(query_list, &query_list_count));
				
				sprintf(gui_result,
								"%lf:%s", RSV_FACTOR, query_display_file);
			}
			else {
				sprintf(gui_result,
								"%lf:EMPTY QUERY", RSV_FACTOR);
			}
		}
		else {
			sprintf(gui_result, "%lf:UNCHANGED", old_rsv_factor);
		}
	}
	else {
		fprintf(hf, ":%d\n", NONE_ASSIGNED);
		sprintf(gui_result, "%d:ERROR", NONE_ASSIGNED);
	}

	fflush(hf);

	Tcl_SetResult(interp, gui_result, TCL_STATIC);
}


int clear_query(ClientData clientData, Tcl_Interp *interp, 
								int argc, char *argv[]) {
	int set = NONE_ASSIGNED;
	int np = NONE_ASSIGNED;
	int window = 1;

  char relspool_file[FILENAME_SIZE];
	char gui_result[SMALL_BUFFER];

  sprintf(relspool_file,
					"/tmp/relspool_file.%d.%d",
					getpid(), relspool_fcount++);
	make_new_bigR_set();
	set_user_rels_flag('N');
	
	fclose(tf);

	hf_header(COMMAND_NO++, "clear_query", PARAMS);
	
  if ((tf = fopen(termset_file, "w")) == NULL) {
		sprintf(ERROR_MESSAGE,
						"EXTRACT: Can't open file '%s' for writing",
						termset_file);
    fprintf(stderr, "%s\n", ERROR_MESSAGE);
		sprintf(gui_result, "ERROR CLEARING QUERY");
		fprintf(hf, ":ERROR\n");
  }
	else {
		fclose(tf);
		
		TERM_NO = 0;
		
		if (USER_RELS_COUNT > 0) {
			/*
			**  Show rels in rels window. First sort into
			**  descending order of document weight
			*/
			
			if (USER_RELS_COUNT > 1) {
				sort_user_rels(user_rels, USER_RELS_COUNT, WGT_SORT);
			}
			
			build_hitlist(set, np, maximum_weight, window,
										relspool_file, &last_relspool_row, relspool_index);
		}
		else {
			sprintf(relspool_file, "NO RELS");
		}
		
		sprintf(gui_result, "QUERY CLEARED OK:%s", relspool_file);
		fprintf(hf, ":OK\n");
	}
	
	Tcl_SetResult(interp, gui_result, TCL_STATIC);

	fflush(hf);
}


int clear_rf(ClientData clientData, Tcl_Interp *interp, 
						 int argc, char *argv[]) {
	int debugging = atoi(getenv("CLEAR_RF_DEBUG"));
	int no_terms;
	int set = NONE_ASSIGNED;
	int np = NONE_ASSIGNED;
	int window = 1;

	char query_display_file[FILENAME_SIZE];
  char relspool_file[FILENAME_SIZE];
	char gui_result[SMALL_BUFFER];
	
  sprintf(relspool_file,
					"/tmp/relspool_file.%d.%d",
					getpid(), relspool_fcount++);
	
	hf_header(COMMAND_NO++, "cleared_rf", NO_PARAMS);
	fflush(hf);

	make_new_bigR_set();
	set_user_rels_flag('N');

	CLEAR_RF = TRUE;

	if (debugging) {
		fprintf(stderr,
						"clear_rf() :: no_terms = %d (TERM_NO = %d)\n",
						no_terms, TERM_NO);
	}	

	WriteUserTerms(termset, TERM_NO);
  TERM_NO = ReadTermsetFile(FALSE);
	no_terms = (TERM_NO > QUERY_SIZE) ? QUERY_SIZE : TERM_NO;
	
	if (debugging) {
		fprintf(stderr,
						"clear_rf() :: no_terms = %d (TERM_NO = %d)\n",
						no_terms, TERM_NO);
	}

	/*
	**  Return the filename containing the query list (for reading
	**  into the text widget. Also put individual lines into the 
	**  character array 'query_list' which is local to this program.
	*/

	sprintf(query_display_file, "%s",
					display_terms(query_list, &query_list_count));
	
	if (debugging) {
		fprintf(stderr,
						"clear_rf() :: query_display_file = [%s]\n",
						query_display_file);
	}

	sprintf(gui_result, "%s", query_display_file);

  if (USER_RELS_COUNT > 0) {
    /*
		**  Show rels in rels window. First sort into
		**  descending order of document weight
		*/
		
    if (USER_RELS_COUNT > 1) {
      sort_user_rels(user_rels, USER_RELS_COUNT, WGT_SORT);
    }
    
    build_hitlist(set, np, maximum_weight, window,
									relspool_file, &last_relspool_row, relspool_index);
	}
	else {
		sprintf(relspool_file, "NO RELS");
	}

	sprintf(gui_result + strlen(gui_result), ":%s", relspool_file);
	Tcl_SetResult(interp, gui_result, TCL_STATIC);
}


int db_search(ClientData clientData, Tcl_Interp *interp, 
							int argc, char *argv[]) {

	int debugging = atoi(getenv("DB_SEARCH_DEBUG"));
  int i;
  int np;

	/*
	**  window = 0 => hitlist widget
	*/
	
  int window = 0;
	
  char hitlist_file[FILENAME_SIZE];
	char gui_result[SMALL_BUFFER];
  
  sprintf(hitlist_file,
					"/tmp/hitlist_file.%d.%d",
					getpid(), hitlist_fcount++);
  
  if (argc > 1) {
    if (NEW_ITERATION) {
      ITERATION_NO++;
      write_terms_to_history_file(TERM_NO, COMMAND_NO++);
      
      bss_search(&np, &maximum_weight);
			
      if (np > 0) {
				/*
				**  return hitlist filename to hitlist_file
				*/
				
				if (debugging) {
					fprintf(stderr, "db_search() :: calling build_hitlist()\n");
				}

				build_hitlist(SHOW_SET, np, maximum_weight, window,
											hitlist_file, &last_hitlist_row, hitlist_index);
				
				if (debugging) {
					fprintf(stderr,
									"db_search() :: hitlist_file = [%s] (%d)\n",
									hitlist_file, strlen(hitlist_file));
				}

				sprintf(gui_result, "%s", hitlist_file);

				if (debugging) {
					fprintf(stderr,
									"db_search() :: gui_result = [%s] (%d)\n", 
									gui_result, strlen(gui_result));
				}
      }
    }
  }
	else {
		printf("Not enough arguments to \"db_search()\"\n");
		sprintf(gui_result, "ERROR: Not enough arguments to \"db_search()\"");
	}

	Tcl_SetResult(interp, gui_result, TCL_STATIC);

	if (debugging) {
		fprintf(stderr,
						"db_search() :: interp->result [%s] (%d)\n", 
						interp->result, strlen(interp->result));
	}
}




/*
**  Makes relevance judgement.
**
**  Called by 'procedures.tcl :: make_rel
**
**  Relevance judgement made on document given by SHOW_DOCID
**  The relj is passed to argv[1][0]
**
**  This function calls 'make_rj.c :: make_rj()' which, according to
**  the relj, calls 'extract.c :: extract_terms()' and adds information
**  to the user_rels[] structure.
**
**  Finally the function calls 'hitlist.c :: build_hitlist()'
**  to update the relspools window.
*/

int make_rel(ClientData clientData, Tcl_Interp *interp,
						 int argc, char *argv[]) {
  /*
  **  Returns from the TCL procedure 'tk_textbox' a value in [P, F, N]
  */

	int debugging = atoi(getenv("MAKE_REL_DEBUG"));
  int i;
	
  int record_no;
  int length;
  int offset;
  int valid_no = TRUE;
  int rj;
	
  int no_terms;
	
  int enable_search = FALSE;
  int optimise = DISABLE_OPTIMISE;

	int old_bigr;

	/*
	**  window = 1 => relspool widget
	*/

	int window = 1;
	int set = NONE_ASSIGNED;
	int np = NONE_ASSIGNED;

	int previous_rel;

  char *in;
  char relj;
  char relspool_file[FILENAME_SIZE];
  char query_display_file[FILENAME_SIZE];

	char gui_result[SMALL_BUFFER];

  sprintf(relspool_file, 
					"/tmp/relspool_file.%d.%d",
					getpid(), relspool_fcount++);
	
  if (argc == 3) {
    sprintf(LAST_COMMAND, "JUDGE");

		/*
		**  Set 'old_bigr' to current value of BIGR
		**
		**  This is to enable users to optimise a query after they
		**  have made a +ve to -ve judgement that makes BIGR fall
		**  below the LR_THRESHOLD.
		*/

		old_bigr = BIGR;
		
    relj = argv[1][0];
		previous_rel = atoi(argv[2]);

		if (debugging) {
			fprintf(stderr, "make_rel() ::         relj = [%c]\n", relj);
			fprintf(stderr, "make_rel() :: Previous_rel = [%d]\n", previous_rel);
		}

    if (relj == 'F') {
      rj = R_FULL;
    }
    else if (relj == 'P') {
      rj = R_PASSAGE;
    }
    else {
      rj = R_NOT;
    }
		
    offset = length = NONE_ASSIGNED;
		
    if (rj == R_FULL) {
      offset = hitlist_index[SHOW_INDEX].offset[1];
      length = hitlist_index[SHOW_INDEX].length[1];
    }
    else if (rj == R_PASSAGE) {
      offset = hitlist_index[SHOW_INDEX].offset[0];
      length = hitlist_index[SHOW_INDEX].length[0];
    }
		
    fprintf(hf, ":%c\n", relj);
    fflush(hf);
		
    /*
    **  make_rj() extracts terms from the document if it is an
    **  F or P judgement or adjusts the termset accordingly
    **  if it is a N rel that has already been seen.
    */

		if (rj == R_NOT || rj == R_FULL || rj == R_PASSAGE) {
			make_rj(SHOW_DOCID, SHOW_DATE, SHOW_SET_RECNO, SHOW_ABS_RECNO,
							rj, offset, length, full_doclen, part_doclen, previous_rel);

			TERM_NO = ReadTermsetFile();

			if (BIGR >= LR_THRESHOLD) {
				expand_query();
				
				sprintf(query_display_file, "%s",
								display_terms(query_list, &query_list_count));
				
				no_terms = (TERM_NO > QUERY_SIZE) ? QUERY_SIZE : TERM_NO;
				enable_search = TRUE;
			}
			else {
				sprintf(query_display_file, "%s",
								display_terms(query_list, &query_list_count));
				
				no_terms = (TERM_NO > QUERY_SIZE) ? QUERY_SIZE : TERM_NO;
				
				enable_search = TRUE;
				
			}

			/*
			**  Show rels in rels window. First sort into
			**  descending order of document weight
			*/
			
			if (BIGR > 0) {
				sort_user_rels(user_rels, USER_RELS_COUNT, WGT_SORT);
			}
			
			last_relspool_row = NONE_ASSIGNED;
			
			build_hitlist(set, np, maximum_weight, window,
										relspool_file, &last_relspool_row, relspool_index);
			
			/*
			**  Don't need to free(relspool_index) since
			**  no memory is malloced for it.
			**
			**  the results returned to make_rel contains 2 flags:
			**
			**    enable_search & optimise
			**
			**  which determine the status of the search & optimise buttons.
			*/

			optimise = FALSE;

			if ((BIGR >= LR_THRESHOLD) && (BIGR > old_bigr)) {
				optimise = TRUE;
			}
			else {
				if ((old_bigr == LR_THRESHOLD) && ((old_bigr - BIGR) == 1)) {
					optimise = TRUE;
				}
			}
			
			enable_search =
				(enable_search == TRUE) ? ENABLE_SEARCH : DISABLE_SEARCH;
			
			sprintf(gui_result,
							"%d:%s:%d:%d:%s",
							no_terms, query_display_file,
							enable_search, optimise, relspool_file);
		}
	}

	Tcl_SetResult(interp, gui_result, TCL_STATIC);
}




int set_query_size(ClientData clientData, Tcl_Interp *interp, 
									 int argc, char *argv[]) {
	int new_query_size;

  char query_display_file[FILENAME_SIZE];
	char gui_result[SMALL_BUFFER];

  hf_header(COMMAND_NO++, "set_query_size", PARAMS);

	if (argc == 2) {
		new_query_size = atoi(argv[1]);
		
    if (new_query_size > MIN_QUERY_SIZE - 1 &&
				new_query_size <= MAX_QUERY_SIZE) {
			QUERY_SIZE = new_query_size;

			if (TERM_NO > 0) {
				sprintf(query_display_file, "%s",
							display_terms(query_list, &query_list_count));
				
				sprintf(gui_result,
								"%d:%s", QUERY_SIZE, query_display_file);
			}
			else {
				sprintf(gui_result,
								"%d:EMPTY QUERY", QUERY_SIZE);
			}
		}
		else {
			/*
			**  Don't change query_size
			*/
			sprintf(gui_result, "%d:UNCHANGED", QUERY_SIZE);
		}
	}
	else {
		/*
		**  Error in call to function - leave QUERY_SIZE alone.
		*/
		sprintf(gui_result, "%d:ERROR", QUERY_SIZE);
	}

	Tcl_SetResult(interp, gui_result, TCL_STATIC);
}


int set_lr_threshold(ClientData clientData, Tcl_Interp *interp, 
										int argc, char *argv[]) {

	int debug = atoi(getenv("SET_LR_THRESHOLD_DEBUG"));

	int new_lr_threshold;

  char query_display_file[FILENAME_SIZE];
	char gui_result[SMALL_BUFFER];

	char env_variable[SMALL_BUFFER];
	char env_value[SMALL_BUFFER];
	char os_command[SMALL_BUFFER];

  hf_header(COMMAND_NO++, "set_lr_threshold", PARAMS);

	if (debug) {
		fprintf(stderr, 
						"set_lr_threshold() :: old threshold = %d\n",
						LR_THRESHOLD);
	}

	if (argc == 2) {
		/*
		**  The new (changed) LR_THRESHOLD is passed to this function
		*/

		new_lr_threshold = atoi(argv[1]);

		fprintf(hf, ":%d\n", new_lr_threshold);

		if (debug) {
			fprintf(stderr, "set_lr_threshold() :: new threshold = %d\n",
							new_lr_threshold);
		}

		if (new_lr_threshold != LR_THRESHOLD) {
			LR_THRESHOLD = new_lr_threshold;

			if (TERM_NO > 0) {
				sprintf(query_display_file, "%s",
								display_terms(query_list, &query_list_count));
				
				sprintf(gui_result,
								"%d:%s", new_lr_threshold, query_display_file);
			}
			else {
				sprintf(gui_result,
								"%d:EMPTY QUERY", new_lr_threshold);
			}
		}
		else {
			sprintf(gui_result, "%d:UNCHANGED", new_lr_threshold);
		}
	}
	else {
		fprintf(hf, ":%d\n", NONE_ASSIGNED);
		sprintf(gui_result, "%d:ERROR", NONE_ASSIGNED);
	}

	fflush(hf);

	Tcl_SetResult(interp, gui_result, TCL_STATIC);
}


int set_rsv_factor(ClientData clientData, Tcl_Interp *interp, 
									 int argc, char *argv[]) {

	int debug = atoi(getenv("SET_RSV_FACTOR_DEBUG"));

	double new_rsv_factor;

  char query_display_file[FILENAME_SIZE];
	char gui_result[SMALL_BUFFER];

  hf_header(COMMAND_NO++, "set_rsv_factor", PARAMS);

	if (debug) {
		fprintf(stderr, 
						"set_rsv_factor() :: old factor = %lf\n",
						RSV_FACTOR);
	}

	if (argc == 2) {
		/*
		**  The new (changed) RSV_FACTOR is passed to this function
		*/

		new_rsv_factor = atof(argv[1]);

		fprintf(hf, ":%lf\n", new_rsv_factor);

		if (debug) {
			fprintf(stderr, "set_rsv_factor() :: new factor = %lf\n",
							new_rsv_factor);
		}

		if (new_rsv_factor != RSV_FACTOR) {
			RSV_FACTOR = new_rsv_factor;

			if (TERM_NO > 0) {
				sprintf(query_display_file, "%s",
								display_terms(query_list, &query_list_count));
				
				sprintf(gui_result,
								"%lf:%s", new_rsv_factor, query_display_file);
			}
			else {
				sprintf(gui_result,
								"%lf:EMPTY QUERY", new_rsv_factor);
			}
		}
		else {
			sprintf(gui_result, "%lf:UNCHANGED", new_rsv_factor);
		}
	}
	else {
		fprintf(hf, ":%d\n", NONE_ASSIGNED);
		sprintf(gui_result, "%d:ERROR", NONE_ASSIGNED);
	}

	fflush(hf);

	Tcl_SetResult(interp, gui_result, TCL_STATIC);
}


int query_expand(ClientData clientData, Tcl_Interp *interp, 
								 int argc, char *argv[]) {
  int no_terms;
	
	char query_display_file[FILENAME_SIZE];
	char gui_result[SMALL_BUFFER];
	
  /*
  **  Display function called in tcl command window
  */
	
  if (strcmp(LAST_COMMAND, "EXPAND") != 0) {
    sprintf(LAST_COMMAND, "EXPAND");
  }
	
  if (EXTRACTED_TERMS) {
    expand_query();
		
    sprintf(query_display_file, "%s",
						display_terms(query_list, &query_list_count));
    
    no_terms = (TERM_NO > QUERY_SIZE) ? QUERY_SIZE : TERM_NO;
		
    sprintf(gui_result, "%d:%s", no_terms, query_display_file);
  }

	Tcl_SetResult(interp, gui_result, TCL_STATIC);
}


int quit_gui(ClientData clientData, Tcl_Interp *interp,
						 int argc, char *argv[]) {
  int no_terms;
  int no_removed_terms;

  int next_rel;
  int term_count = 0;
	
  char temp_file[FILENAME_SIZE];
  char temp[MAX_TERM_LENGTH];
	
  char elapsed[RUNTIME_LENGTH];
	char gui_result[SMALL_BUFFER];
	
  /*
  **  Write user rels to rels_file.
  */
	
  show_time("Quit_Gui", START);
	
  /*
  **  Update ELAPSED_TIME
  */
	
  current_time(elapsed);
	
  hf_header(COMMAND_NO++, "quit", PARAMS);
  fprintf(hf, ":%d\n", ITERATION_NO);
  fclose(hf);
	
  RemoveTempfiles(getpid());
  shut_down();
  show_time("Quit_Gui", FINISH);
  exit(0);
}


int reinstate_term(ClientData clientData, Tcl_Interp *interp, 
									 int argc, char *argv[]) {
  /*
  **  Result returned to TCL procedure 'reinstate_term' is of the form:
  **
  **    <query_display_file>:<remove_filename>
  **
  **  where <remove_filename> = "EMPTY" if there are none left in the list
  */
	
  int next_term = 0;
  int no_terms;
  int row;

	char gui_result[SMALL_BUFFER];
	
  char term[MAX_PHRASE_LENGTH];
  char temp[MAX_TERM_LENGTH];
  char query_display_file[FILENAME_SIZE];
	
  char removed_terms_file[FILENAME_SIZE];

  FILE *rtf;
	
  sprintf(removed_terms_file, "%s/%s.%d",
					Temp_dir, REMOVED_TERMS_FILE, getpid());
	
  show_time("ReinstateTerm", START);
	
  switch (argc) {
	  case 2 :
			/*
      **  Valid remove statement.
      */
			
      sprintf(LAST_COMMAND, "REMOVE");
      row = atoi(argv[1]) - 1;
			
      /*
      **  Display query in tcl command window
      */
			
      RestoreTerm(removed_terms[row].term);
      no_terms = (TERM_NO > QUERY_SIZE) ? QUERY_SIZE : TERM_NO;
      sprintf(query_display_file, "%s",
							display_terms(query_list, &query_list_count));
			
      switch (query_list_count) {
        case 0 :
					sprintf(gui_result, "EMPTY:");
					break;
			default :
				sprintf(gui_result, "%s:", query_display_file);
				break;
      }
			
      NO_REMOVED_TERMS--;
			
      if (NO_REMOVED_TERMS == 0) {
				sprintf(gui_result + strlen(gui_result), "EMPTY");
      }
      else {
				sprintf(gui_result + strlen(gui_result),
								"%s", removed_terms_file);
      }
			
      break;
	  default :
      /*
      **  Error in remove statement
      */
      sprintf(gui_result, "## RESTORE TERM ERROR ##\n");
      break;
  }
	
  /*
  **  Empty removed_terms structure
  */
	
  clear_structure(removed_terms);
	
  /*
  **  Now re-read removed_terms_file into removed_terms[]
  **  if there are any left.
  */
	
  if (NO_REMOVED_TERMS > 0) {
		next_term=0;
    rtf = fopen(removed_terms_file, "r");
    
    while (!feof(rtf)) {
      if (fgets(temp, MAX_TERM_LENGTH, rtf) == NULL) {
				break;
      }
      else {
				temp[strlen(temp) - 1] = '\0';
				sprintf(removed_terms[next_term++].term, "%s", temp);
      }
    }
		
    NO_REMOVED_TERMS = next_term;
    fclose(rtf);
  }
	
	Tcl_SetResult(interp, gui_result, TCL_STATIC);

  show_time("ReinstateTerm", FINISH);
}


int remove_term(ClientData clientData, Tcl_Interp *interp, 
								int argc, char *argv[]) {
  /*
  **  Result returned to TCL procedure 'remove_term' is of the form:
  **
  **    <query_display_file>:<remove_filename>
  **
  **  where <remove_filename> = "EMPTY" if none removed.
  */

  int next_term;
  int no_terms;
  int row;

	char gui_result[SMALL_BUFFER];

  char term[MAX_PHRASE_LENGTH];
  char query_display_file[FILENAME_SIZE];

  char removed_terms_file[FILENAME_SIZE];

  FILE *rtf;

  show_time("RemoveTerm", START);

  sprintf(removed_terms_file, "%s/%s.%d",
					Temp_dir, REMOVED_TERMS_FILE, getpid());

  switch (argc) {
    case 2 :
      /*
      **  Valid remove statement.
			**  'C' array items are numbered from 0.
			**  Tcl text rows are numbered from 1
      */

      sprintf(LAST_COMMAND, "REMOVE");
      row = atoi(argv[1]) - 1;
			
      /*
      **  Display query in tcl command window
      */
			
      DeleteTerm(query_list[row].term);
			
      sprintf(removed_terms[NO_REMOVED_TERMS].term, "%s",
							query_list[row].term);
			
      NO_REMOVED_TERMS++;
			
      /*
      **  Append term to removed_terms_file
      */
			
      rtf = fopen(removed_terms_file, "a");
      fprintf(rtf, "%s\n", query_list[row].term);
			
      fclose(rtf);
			
      no_terms = (TERM_NO > QUERY_SIZE) ? QUERY_SIZE : TERM_NO;
      sprintf(query_display_file, "%s",
							display_terms(query_list, &query_list_count));
			
      switch (query_list_count) {
			  case 0 :
					sprintf(gui_result, "EMPTY:");
					break;
		  	default :
					sprintf(gui_result, "%s:", query_display_file);
					break;
      }

      sprintf(gui_result + strlen(gui_result), "%s",
							removed_terms_file);

      break;
    default :
      /*
      **  Error in remove statement
      */
      sprintf(gui_result, "## TERM REMOVAL ERROR ##\n");
      break;
  }
	
	Tcl_SetResult(interp, gui_result, TCL_STATIC);

  show_time("RemoveTerm", FINISH);
}


int show_rec(ClientData clientData, Tcl_Interp *interp, 
						 int argc, char *argv[]) {
  int current_rel;
  int i;
  int next_record = NONE_ASSIGNED;
	int last_record = atoi(getenv("MAX_RECS_TO_SHOW"));
  int r;
  int set_recno;
  int row;
  int length;
  int doc_part = NONE_ASSIGNED;
  int valid_no = TRUE;

	char gui_result[SMALL_BUFFER];

	char dummy[FILENAME_SIZE + 20];
	char final_show_file[FILENAME_SIZE];

	char para_file;

	if (params_file.para_file) {
		para_file = 'P';
	}
	else {
		para_file = 'N';
	}

  switch (argc) {
    case 2 :
      sprintf(LAST_COMMAND, "SHOW");

      /*
      **  A valid show command -- argv[1] contains the number of the
      **  row clicked in the hitlist. set row to atoi(argv[1]) - 1
      **  so that an index in the range 0 to number_of_hitlist_lines
      **  is generated.
      */
			
      row = atoi(argv[1]) - 1;
      SHOW_SET_RECNO = NONE_ASSIGNED;
      
      if ((row >= 0) && (row < last_hitlist_row)) {
				for (i = 0; i < last_record; i++) {
					if ((row >= hitlist_index[i].first_row) &&
							(row <= hitlist_index[i].last_row)) {
						
						SHOW_SET_RECNO = hitlist_index[i].set_recno;
						SHOW_ABS_RECNO = hitlist_index[i].abs_recno;
						SHOW_WGT = hitlist_index[i].weight;
						sprintf(SHOW_DOCID, "%s", hitlist_index[i].docid);
						SHOW_INDEX = i;
						
						/*
						**  Set next_record to the first row of the next doc
						*/
						next_record = hitlist_index[i + 1].first_row;
						
						break;
					}
				}
      }
			
      if (SHOW_SET_RECNO != NONE_ASSIGNED) {
				/*
				**  Display query in tcl command window
				**
				**  First get SHOW_DOCID
				**
				**  Now check to see if it has already been seen to return
				**  any previous relevance judgement in 'current_rel'
				*/
				
				current_rel = R_NONE;

				if (USER_RELS_COUNT > 0) {
					for (r = 0; r < USER_RELS_COUNT; r++) {
						if (strcasecmp(SHOW_DOCID, (user_rels + r)->docid) == 0) {
							current_rel = (user_rels + r)->rj;
							break;
						}
					}
				}
				
				/*
				**  Return string to tcl procedure 'show_rec' of the form:
				**
				**    <curent_rel>:<docid>:<show_filename>:
				**    <F | P>:
				**    <next_record in sequence>:
				**    <the document part to move cursor to>
				**    <P|N> (whether a passage search or not)
				**
				**  Return full_doclen & part_doclen to be passed (via
				**  make_rj() to extract_terms().
				*/

				show_record(SHOW_SET, SHOW_SET_RECNO, maximum_weight,
										&doc_part, final_show_file,
										&full_doclen, &part_doclen);

				sprintf(gui_result, "%d:%s:%s:%d:%d:%c",
								current_rel, SHOW_DOCID, final_show_file,
								next_record, doc_part, para_file);
      }
      else {
				sprintf(gui_result, "NONE_FOUND");
      }
      break;
	  default :
      /*
      **  An invalid show command
      */
      sprintf(gui_result, "## SHOW COMMAND ERROR ##\n");
      break;
  }

	Tcl_SetResult(interp, gui_result, TCL_STATIC);
}


int show_rels(ClientData clientData, Tcl_Interp *interp, 
							int argc, char *argv[]) {
	int set = NONE_ASSIGNED;
	int np = NONE_ASSIGNED;
	int window = 1;

	char gui_result[SMALL_BUFFER];

  char relspool_file[FILENAME_SIZE];
	
  sprintf(relspool_file,
					"/tmp/relspool_file.%d.%d",
					getpid(), relspool_fcount++);
	
  if (BIGR > 0) {
    /*
		**  Show rels in rels window. First sort into
		**  descending order of document weight
		*/
		
    if (USER_RELS_COUNT > 1) {
      sort_user_rels(user_rels, USER_RELS_COUNT, WGT_SORT);
    }
		
    
    build_hitlist(set, np, maximum_weight, window,
									relspool_file, &last_relspool_row, relspool_index);
		
    sprintf(gui_result, "%s", relspool_file);
  }
  else {
    sprintf(gui_result, "EMPTY");
    relspool_fcount--;
  }

	Tcl_SetResult(interp, gui_result, TCL_STATIC);
}


int term_entry(ClientData clientData, Tcl_Interp *interp, 
							 int argc, char *argv[]) {
	int debugging = atoi(getenv("TERM_ENTRY_DEBUG"));
  int i;
	
  int last_term_no = TERM_NO;
  int no_terms;
  int no_removed_terms;
	
  char query[MAX_PHRASE_LENGTH];
  char terms_not_found[2 * SMALL_BUFFER];
  char query_display_file[FILENAME_SIZE];
	char gui_result[SMALL_BUFFER];
	
  if (argc == 2) {
		EDITING_TOPIC = TRUE;
    sprintf(LAST_COMMAND, "INPUT");
		
    sprintf(query, "%s", argv[1]);
		
    /*
    **  Send back a string to TCL proc 'term_entry' structured:
    **
    **     <nterms>:<entry_type>:<query_fname>:<removed_fname>:<extra>
    **
    **        <entry_type> in [S(ome found), N(one found), E(rror)]
    **       <query_fname> is the name of the file containing
    **                     the query terms
    **     <removed_fname> is the name of the file containing
    **                     any removed/dropped out terms
    **             <extra> is a string containg a list of not found
    **                     or already entered terms, or an error message
    */
		
    if (strlen(query) > 0) {
      term_input(query, terms_not_found);
			
			if (strcmp(terms_not_found, "TOO LONG") == 0) {
				/*
				**  Input too long
				*/
				sprintf(gui_result, "%s", terms_not_found);
			}
			else {
				no_terms = (TERM_NO > QUERY_SIZE) ? QUERY_SIZE : TERM_NO;
				
				/*
				**  Return the filename containing the query list (for reading
				**  into the text widget. Also put individual lines into the 
				**  character array 'query_list' which is local to this program.
				*/
				sprintf(query_display_file, "%s",
								display_terms(query_list, &query_list_count));
				
				/*
				**  query_display_file will be set to the name of the file to be
				**  loaded into the text widget or EMPTY if none.
				*/
			
				sprintf(gui_result, "%s", query_display_file);
				
				if (strlen(terms_not_found) != 0) {
					/*
					**  One or more of the terms were not found or already
					**  exist in the termset under another source (i.e. the
					**  parsed forms of the terms were identical).
					*/
					sprintf(gui_result + strlen(gui_result),
									":%s", terms_not_found);
				}
				else {
					sprintf(gui_result + strlen(gui_result), ":EMPTY");
				}
			}
    }
    else {
      sprintf(gui_result, "EMPTY:Probably entry is an empty string.\n");
    }
  }
  else {
    sprintf(gui_result, "EMPTY:No query entered.\n");
  }

	Tcl_SetResult(interp, gui_result, TCL_STATIC);

	if (debugging) {
		fprintf(stderr,
						"term_entry() :: gui_result = [%s]\n",
						interp->result);
	}
}


int window_title(ClientData clientData, Tcl_Interp *interp, 
								 int argc, char *argv[]) {
  char dots[40];
	char gui_result[SMALL_BUFFER];

  sprintf(dots, "........................................");
	
  sprintf(gui_result,
					"PID: [%d] ..... Current logfile: [%s/%s]",
					getpid(), getenv("OKAPI_LOGS_DIR"),	TOPIC_NAME);

	Tcl_SetResult(interp, gui_result, TCL_STATIC);
}


