/*
**  initialise.c
*/

#define G_EXTERN extern

#include "defs.h"

void setup_filenames();

G_EXTERN void initialise_global_variables();
G_EXTERN FILE *OpenFile(char *, char *);
G_EXTERN int open_database(char *, int);
G_EXTERN void make_RS_sets(int);

void cleanabort()
{
  abort();
}

void set_error_file()
{
  extern FILE *Err_file;

	Err_file = fopen("/dev/null", "w");
}


void initialisation()
{
	char *DISPLAY_VAR = NULL;

  /*
  **  Initialise the BSS
  */

  iinit();

	set_error_file();

  /*
  **  Initialise global variables.
  **
  **  HIT_COUNT & SHOW_COUNT are variables to increment hitlist
  **  and fullrecs files
  */

	initialise_global_variables();

  /*
  **  set up filenames.
  **
  **  a) logfiles
  **
  **  All log files will be kept in <OKAPI_ROOT>/OkapiLogs
  **
  **  First make this directory if it does not exist
  */

  setup_filenames();

	fprintf(stderr, "File [%s] ..... ", history_file);

	hf = OpenFile(history_file, "w");

	fprintf(stderr, "OPENED.\n");

  /*
  **  Open default database if not already open
  */

	DB_OPEN = open_database(DATABASE_NAME, FALSE);

	make_RS_sets(FALSE);
	fprintf(stderr, "OKAPI INITIALISATION COMPLETE.\n");
}


void setup_filenames()
{
	/*
	**  The first three logs contain:
	**
	**  a) a complete command/result history for a search.
	**  b) the entire set of user-entered and extracted terms at
	**     the current stage in the search.
	**  c) the complete set of positive relevance judgements at
	**     the current stage in the search.
	*/

  sprintf(history_file,
					"%s/%s/%s",
					getenv("OKAPI_LOGS_DIR"), TOPIC_NAME, HISTORY_FILE);

  sprintf(termset_file, 
					"%s/%s/%s",
					getenv("OKAPI_LOGS_DIR"), TOPIC_NAME,	TERMSET_FILE);

  sprintf(rels_file, 
					"%s/%s/%s", 
					getenv("OKAPI_LOGS_DIR"), TOPIC_NAME,	RELS_FILE);
	
  /*
  **  The following files are all working files that will be
	**  deleted at the end of the search.
	**
	**  d) Working query file.
	**  e) All terms that have appeared in the query at any iteration
	**  f) The current iteration terms
	**  g) The current hitlist 
  **  
	**  For documents that are selected to be seen in full.
	**
	**  h) The full document   and the removed terms display.
	**  g) The part of the document to be show (might be same as (h).
  */
	
  sprintf(working_query_file, "%s/%s.%d",
					Temp_dir, WORKING_QUERY_FILE, getpid());
  sprintf(all_iterations_file, "%s/%s.%d",
					Temp_dir, ALL_ITERATIONS_FILE, getpid());
  sprintf(current_iteration_file, "%s/%s.%d",
					Temp_dir, CURRENT_ITERATION_FILE, getpid());
	sprintf(show_file, "%s/%s.%d",
					Temp_dir, SHOW_FILE, getpid());
	sprintf(hitlist_file,	"%s/%s.%d",
					Temp_dir, HITLIST_FILE, getpid());
	sprintf(fullrecs_file, "%s/%s.%d",
					Temp_dir, FULL_RECS_FILE, getpid());

}
