/*
**  interface.cc
**
**  mg - 7th June 1995
*/

#define G_EXTERN extern

#include "defs.h"
#include "GLOB.h"

G_EXTERN int GetAndCheckVars();
G_EXTERN int Do_AppDeConstruct();

G_EXTERN "C" {
  void exit();
  int isatty();
  int read();
  char *strrchr();
}

/*
**  The main window for the application.
**  If NULL then the application no longer exists
*/

static Tk_Window enqWD;

/*
**  Used to assemble lines of terminal input into Tcl tclcmds
*/

static Tcl_DString tclcmd;

/*
**  Non-zero means standard input is a terminal-like device.
**  Zero means it's a file.
*/

static int tty;

static char errorExitCmd[] = "exit 1";

/*
** Command-line options:
*/

static int synchronize = 0;
static char *fileName = NULL;
static char *name = NULL;
static char *display = NULL;
static char *geometry = NULL;
static char *db_name = NULL;

static Tk_ArgvInfo argTable[] = {
  {
    "-file", TK_ARGV_STRING, (char *) NULL, (char *) &fileName,
    "File from which to read commands"
  },
  {
    "-geometry", TK_ARGV_STRING, (char *) NULL, (char *) &geometry,
    "Initial geometry for window"
  },
  {
    "-display", TK_ARGV_STRING, (char *) NULL, (char *) &display,
    "Display to use"
  },
  {
    "-name", TK_ARGV_STRING, (char *) NULL, (char *) &name,
    "Name to use for application"
  },
  {
    "-db", TK_ARGV_CONSTANT, (char *) 1, (char *) &db_name,
    "Use synchronous mode for display server"
  },
  {
    "-sync", TK_ARGV_CONSTANT, (char *) 1, (char *) &synchronize,
    "Use synchronous mode for display server"
  },
  {
    (char *) NULL, TK_ARGV_END, (char *) NULL, (char *) NULL,
    (char *) NULL
  }
};

/*
**  Forward declarations for procedures defined later in this file:
*/

static void Prompt _ANSI_ARGS_((Tcl_Interp *interp, int partial));
static void StdinProc _ANSI_ARGS_((ClientData clientData, int mask));
G_EXTERN int OKAPI_TclInit(Tcl_Interp *theInterp);
G_EXTERN void Enquire_LogClient(char *);

/*
**  Error message functions.
**
**  1. Incorrect program call
*/

void usage(char progname[]) {
	printf("\n\n**\n");
	printf("**\n");
	printf("**  Incorrect call of program\n");
	printf("**\n");
  printf("**  Please rerun as:\n");
	printf("**\n");
  printf("**  %s <topic_no> <user_code> <relevance_feedback>\n",
				 progname);
	printf("**  Where:\n\n");
	printf("**    <topic_no> is greater than 0\n");
	printf("**    <user_code> is an identifier for the searcher.\n");
	printf("**    <relevance_feedback> indicates whether you want\n");
	printf("**    to have relevance feedback or not.\n\n");
	printf("**    <relevance_feedback> IN [ 0 | 1 ] => [ No RF | RF ]\n");
	printf("**\n");
  exit(0);
}

/*
**  2. No DISPLAY variable set
*/

void no_display() {
	printf("\n\n*******************************************************\n");
	printf("**                                                   **\n");
	printf("**  There is no DISPLAY environment variable set.    **\n");
	printf("**                                                   **\n");
	printf("**  Please set it as, e.g.                           **\n");
	printf("**                                                   **\n");
	printf("**       setenv DISPLAY pace.is.ac.uk:0.0            **\n");
	printf("**                                                   **\n");
	printf("**  and then re-run the program.                     **\n");
	printf("**                                                   **\n");
	printf("*******************************************************\n\n");
}

/*
**  No file containing valid database names: <$GUI_CONFIG_FILES>/.okapi_db
*/

void no_dbname_file() {
	printf("\n\n");
	printf("**\n");
	printf("**      There is no file \".okapi_db\" in\n\n");
	printf("**      [%s]\n\n", getenv("GUI_CONFIG_FILES"));
	printf("**      This file must exist and the first line\n");
	printf("**      must contain the name of the database\n\n");
	exit(0);
}

/*
**  Create logfile sub-directory
*/

void create_userdir(char *program, char userid[]) {
	int buffer_size;

	int next_log;

	int run_no = 0;
	int dir_exists = FALSE;
	int next_dir;
	int no_entries;

	char topic_dirname[FILENAME_SIZE];
	char full_entry[FILENAME_SIZE];

	struct logfile_info *topic_runs;

	/*
	**  First read directory entries into dirname[][]
	**  There will be <no_entries> entries.
	*/

	buffer_size = MAX_NO_TOPICS * sizeof(struct logfile_info);

	if ((topic_runs = (struct logfile_info *) malloc(buffer_size)) == NULL) {
		malloc_error(program, "create_userdir", "topic_runs", buffer_size);
	}
	
	no_entries = read_topic_names(topic_runs);

	for ( ; ; ) {
		/*
		**  Infinite loop - keep going until we find
		**  a <topic_dirname> that doesn't exist.
		*/
		sprintf(TOPIC_NAME, "t%d.%s.%d", TOPIC_NO, userid, run_no);
		sprintf(topic_dirname, "%s/%s",	getenv("OKAPI_LOGS_DIR"), TOPIC_NAME);

		/*
		**  Check against current entries.
		**  First set dir_exists to FALSE
		*/

		dir_exists = FALSE;

		for (next_dir = 0; next_dir < no_entries; next_dir++) {
			if (strcmp((topic_runs + next_dir)->logfile, TOPIC_NAME) == 0) {
				/*
				**  Already exists
				*/
				dir_exists = TRUE;
				break;
			}
		}

		if (!dir_exists) {
			/*
			**  If it doesn't exist create it.
			*/

			sprintf(OS_COMMAND,	"mkdir %s", topic_dirname);
		
			if (system(OS_COMMAND) < 0) {
				perror(OS_COMMAND);
			}
			else {
				printf("*******************************************************\n\n");
				printf("    [%s] successful\n\n", OS_COMMAND);
				printf("*******************************************************\n\n");
			}
			break;
		}
		run_no++;
	}

	free(topic_runs);
}


/*
**  Main program
*/

int main(int argc, char *argv[]) {
  int debugging = atoi(getenv("MAIN_DEBUG"));
  int max_no_terms;
  int max_no_rels;
  int buffer_size;
  
	int run_no = 0;

	int topic;
  int code;
  int i;

	int feedback_param = TRUE;
	
  char *args;
  char *parameter;
  char *message;
  char buf[20];
	
  char filename[FILENAME_SIZE];
	char dbname[MAX_DBNAME_LENGTH];
	char temp[MAX_DBNAME_LENGTH];
	char userid[32];

	int no_lines;
	int next_entry = 0;
	char init_file[V_SMALL_BUFFER][V_SMALL_BUFFER];

  FILE *fp;

	/*
	**  Check that the DISPLAY variable has been set.
	**  If not display error message and terminate the program.
	*/
	
	if (getenv("DISPLAY") == NULL) {
		no_display();
		exit(0);
	}
	
  theInterp = Tcl_CreateInterp();
	
#ifdef TCL_MEM_DEBUG
  Tcl_InitMemory(theInterp);
#endif
	
  /*
  **  Parse command-line arguments.
  */
	
  /*
  **  Thien's code for selecting initialisation parameters
  **  from the command line arguments.
  */
	
  DB_OPEN = FALSE;
	
	if (argc != 4) {
		usage(argv[0]);
	}
	else {
		TOPIC_NO = atoi(argv[1]);

		if (TOPIC_NO <= 0) {
			usage(argv[0]);
		}

		sprintf(userid, "%s", argv[2]);

		feedback_param = atoi(argv[3]);

		switch (feedback_param) {
		  case 0 : case 1 :
				FEEDBACK = feedback_param;
				break;
		  default :
				usage(argv[0]);
				break;
		}
	}

	/*
	**  First set up name of OKAPI_LOGS_DIR and create it if necessary.
	**
	**  1. Create OKAPI_LOGS_DIR
	*/
	
	sprintf(OS_COMMAND, "mkdir %s", getenv("OKAPI_LOGS_DIR"));
	system(OS_COMMAND);

	/*
	**  Set up default parameters:
	**
	**    TOPIC_COUNT, NO_OF_TOPICS and TOPIC_NAME.
	**
	**  Store TOPIC_NAME TOPIC_LIST[0].
	**
	**  Profiles are numbered from 1 .. MAX_NO_TOPICS.
	**  Hence any loops will go from 1 TO (MAX_NO_TOPICS + 1)
	**  NO_OF_TOPICS is always 1 more than the number of topics.
	*/
	
	TOPIC_COUNT = 0;
	NO_OF_TOPICS = 1;
	
	/*
	**  Make a default TOPIC_NAME and create a subdirectory
	**  of the same name in $HOME/LOGS_DIR
	*/

	create_userdir(argv[0], userid);

  printf("Okapi Initialisation\n");
	printf("====================\n\n");
	
  /*
  **  If a display was specified, put it into the DISPLAY
  **  environment variable so that it will be available for
  **  any sub-processes created by us.
  */

  if (display != NULL) {
    Tcl_SetVar2(theInterp, "env", "DISPLAY", display, TCL_GLOBAL_ONLY);
  }
	
  /*
  **  Initialize the Tk application.
  */

  enqWD = TkCreateMainWindow(theInterp, display, "Okapi", "Tk");

  if (enqWD == NULL) {
    fprintf(stderr, "%s\n", theInterp->result);
    exit(1);
  }
	
  if (synchronize) {
    XSynchronize(Tk_Display(enqWD), True);
  }
	
  Tk_GeometryRequest(enqWD, 800, 750);
	
  /*
  **  Make command-line arguments available in the Tcl variables
  **  "argc" and "argv".  Also set the "geometry" variable from
  **  the geometry specified on the command line.
  */

  args = Tcl_Merge(argc - 1, argv + 1);
  Tcl_SetVar(theInterp, "argv", args, TCL_GLOBAL_ONLY);
  ckfree(args);
  sprintf(buf, "%d", argc-1);
	
  Tcl_SetVar(theInterp, "argc", buf, TCL_GLOBAL_ONLY);
  Tcl_SetVar(theInterp, "argv0",
						 (fileName != NULL) ? fileName : argv[0], TCL_GLOBAL_ONLY);
	
  if (geometry != NULL) {
    Tcl_SetVar(theInterp, "geometry", geometry, TCL_GLOBAL_ONLY);
  }
	
  /*
  **  Set the "tcl_interactive" variable.
  */
	
  tty = isatty(0);
  Tcl_SetVar(theInterp, "tcl_interactive",
						 ((fileName == NULL) && tty) ? "1" : "0", TCL_GLOBAL_ONLY);
	
  /*
  **  Invoke application-specific initialization.
  */
	
  if (OKAPI_TclInit(theInterp) != TCL_OK) {
    fprintf(stderr, "Tcl_AppInit failed: %s\n", theInterp->result);
  }

  initialisation();

  /*
  ** Invoke the script specified on the command line, if any.
  */
	
  if (fileName != NULL) {
		if (code != TCL_OK) {
			goto error;
    }
    else {
      printf("theInterp->result [%s]\n", theInterp->result);
    }
		
    tty = 0;
  }
  else {
    /*
    **  Commands will come from standard input, so set up an event
    **  handler for standard input.  If the input device is a file
		**  evaluate the .rc file
		**
		**  if one has been specified, set up an event handler for standard
		**  input, and print a prompt if the input device is a terminal.
    */

    if (tcl_RcFile != NULL) {
      char *fullName;
			FILE *f;

			fullName = tcl_RcFile;

			if (debugging) {
				fprintf(stderr,
								"interface.cc :: main() - fullName = [%s]\n", fullName);
			}
			
      if (fullName == NULL) {
				fprintf(stderr, "%s\n", theInterp->result);
      }
      else {
				/*
				**  First rewrite first line of .okapi_rc to set RF_FLAG
				*/

				/*
				buffer_size = V_SMALL_BUFFER * (V_SMALL_BUFFER + 1);

			  if ((init_file = (char *) malloc(buffer_size)) == NULL) {
					malloc_error("interface.cc :: main()", "main()",
											 "init_file", buffer_size);
				}
				*/

				if ((f = fopen(fullName, "r")) != NULL) {
					next_entry = 0;
					
					while (!feof(f)) {
						if (fgets(temp, V_SMALL_BUFFER, f) == NULL) {
							break;
						}
						else {
							temp[strlen(temp) - 1] = '\0';
							/*
							**  Need to rewrite the line that sets the
							**  the global variable <rf_flag>
							*/
							
							if (strncmp(temp, "set rf_flag", 11) == 0) {
								sprintf(init_file[next_entry++],
												"set rf_flag %d", FEEDBACK);
							}
							else {
								sprintf(init_file[next_entry++], "%s", temp);
							}
						}
					}
					
					fclose(f);
					
					no_lines = next_entry;

					if ((f = fopen(fullName, "w")) != NULL) {
						for (next_entry = 0; next_entry < no_lines; next_entry++) {
							fprintf(f, "%s\n", init_file[next_entry]);

							if (debugging) {
								fprintf(stderr,
												"init_file[%d] = [%s]\n",
												next_entry, init_file[next_entry]);
							}
						}
						
						fclose(f);
					}
				}

				if ((f = fopen(fullName, "r")) != NULL) {
					code = Tcl_EvalFile(theInterp, fullName);
					
					if (code != TCL_OK) {
						fprintf(stderr, "%s\n", theInterp->result);
					}
					else {
						fprintf(stderr, "%s\n", theInterp->result);
					}
					
					fclose(f);
				}

				/*
				**  free(init_file);
				*/
      }
    }
		
    Tk_CreateFileHandler(0, TK_READABLE, StdinProc, (ClientData) 0);
		
    if (tty) {
      Prompt(theInterp, 0);
    }
  }
	
  fflush(stdout);
  Tcl_DStringInit(&tclcmd);
	
	/*
	**  Allocate memory for termset and user_rels structures.
	**
	**  .okapi_rc has been read now so can read environment
	**  variables MAX_TERMSET_SIZE & MAX_RELS
	*/

	max_no_terms =  atoi(getenv("MAX_TERMSET_SIZE"));
  buffer_size = max_no_terms * sizeof(struct query_terms);
	
  if ((termset = (struct query_terms *) malloc(buffer_size)) == NULL) {
    malloc_error("tcl_interface.cc", "main()", "termset", buffer_size);
  }
  else {
		fprintf(stderr,
						"\n\nMEMORY ALLOCATIONS:   termset:%6d Kbytes ",
						buffer_size/1024);
		fprintf(stderr, "(maximum %d terms)\n", max_no_terms);
	}
  
	max_no_rels =  atoi(getenv("MAX_RELS"));
  buffer_size = max_no_rels * sizeof(struct user_rj);
	
  if ((user_rels = (struct user_rj *) malloc(buffer_size)) == NULL) {
    malloc_error("tcl_interface.cc", "main()", "user_rels", buffer_size);
  }
  else {
		fprintf(stderr,
						"                    user_rels:%6d Kbytes ",
						buffer_size/1024);

		fprintf(stderr, "(maximum %d documents)\n",	max_no_rels);
	}
  
  /*
  **  Should be able to set main window geometry here using:
	**
	**   Tk_MoveWindow(enqWD,
                atoi(getenv("OFFSET_X")),	atoi(getenv("OFFSET_Y")));
	**
	**  where OFFSET_X, OFFSET_Y are set in ".okapi_rc". However, these
	**  won't work here so still set main window geometry in ".okapi_rc"
	**  but set the window geometry when the widgets are packed in
	**  procedures.tcl :: pack_it_all{ }.
	**
	**  Loop infinitely, waiting for commands to execute.
	**  When there are no windows left, Tk_MainLoop returns and we exit.
	*/

  if (GetAndCheckVars() == TCL_OK) {
    Tk_MainLoop();
  }

  /*
  **  Don't exit directly, but rather invoke the Tcl "exit" command.
  **  This gives the application the opportunity to redefine "exit"
  **  to do additional cleanup.
  */

  Tcl_Eval(theInterp, "exit");
  exit(1);

error:
  message = Tcl_GetVar(theInterp, "errorInfo", TCL_GLOBAL_ONLY);

  if (message == NULL) {
    message = theInterp->result;
  }

  fprintf(stderr, "%s\n", message);
  Tcl_Eval(theInterp, errorExitCmd);

  /*
  **  Need 'return 1' only to prevent compiler warnings.
  */
  return 1;
}


int GetAndCheckVars ()
{
  /*
  **  GetAndCheckVars() --
  **  
  **  Needed Vars to communicate between the Internal and External works
  */

  return TCL_OK;
}

static void StdinProc(ClientData clientData, int mask) /* Not used. */
{
  /*
  **  StdinProc --
  **
  **  This procedure is invoked by the event dispatcher whenever standard
	**  input becomes readable. It grabs the next line of input characters,
	**  adds them to a command being assembled, and executes the command if
	**  it's complete.
  **
  **  Results:  NONE.
  **
  **  Side effects:  Could be almost arbitrary, depending on command typed.
  */

  char input[TCL_BUFFER_SIZE];
  static int gotPartial = 0;
  char *cmd;
  int code, count;
  
  count = read(fileno(stdin), input, TCL_BUFFER_SIZE);
  
  if (count <= 0) {
    if (!gotPartial) {
      if (tty) {
				Tcl_Eval(theInterp, "exit");
				exit(1);
      }
      else {
				Tk_DeleteFileHandler(0);
      }
      return;
    }
    else {
      count = 0;
    }
  }

  cmd = Tcl_DStringAppend(&tclcmd, input, count);

  if (count != 0) {
    if ((input[count-1] != '\n') && (input[count-1] != ';')) {
      gotPartial = 1;
      goto prompt;
    }

    if (!Tcl_CommandComplete(cmd)) {
      gotPartial = 1;
      goto prompt;
    }
  }

  gotPartial = 0;
  
  /*
  **  Disable the stdin file handler while evaluating the command;
  **  otherwise if the command re-enters the event loop we might
  **  process commands from stdin before the current command is
  **  finished.  Among other things, this will trash the text of the
  **  command being evaluated.
  */
  
  Tk_CreateFileHandler(0, 0, StdinProc, (ClientData) 0);
  
  if (strncmp(cmd, "exit", 4)== 0) {
    Do_AppDeConstruct();
  }

  code = Tcl_RecordAndEval(theInterp, cmd, 0);
  Tk_CreateFileHandler(0, TK_READABLE, StdinProc, (ClientData) 0);
  Tcl_DStringFree(&tclcmd);

  if (*theInterp->result != 0) {
    if ((code != TCL_OK) || (tty)) {
      printf("%s\n", theInterp->result);
    }
  }

  /*
  **  Output a prompt.
  */
  
 prompt:
  if (tty) {
    Prompt(theInterp, gotPartial);
  }
}



static void Prompt(Tcl_Interp *interp, int partial) {
  /*
  **  Issue a prompt on standard output, or invoke a script to
	**  issue the prompt.
  **
  **  Results:  NONE.
  **
  **  Side effects:  A prompt gets output, and a Tcl script
  **                 may be evaluated in interp.
  */

  char *promptCmd;
  int code;
  
  promptCmd = Tcl_GetVar(interp, partial ? "tcl_prompt2" :
			 "tcl_prompt1", TCL_GLOBAL_ONLY);
  
  if (promptCmd == NULL) {
  defaultPrompt:
    if (!partial) {
      fputs("% ", stdout);
    }
  }
  else {
    code = Tcl_Eval(interp, promptCmd);

    if (code != TCL_OK) {
      Tcl_AddErrorInfo(interp,
		       "\n    (script that generates prompt)");
      fprintf(stderr, "%s\n", interp->result);
      goto defaultPrompt;
    }
  }
  fflush(stdout);
}




