/*
**  search.c
**
**  mg July 1995
**
**  Generates a document set from the working query, as written to the
**  working query file in the function display_terms(). The contents of
**  the working query file are written to the current_iteration_file
**  for use when establishibg the average rsv of the new working query.
**
**  Modified to incorporate passage retrieval. See:
**
**     ~okapi/src/bss/interface4/DOCS/passage.retrieval
**
**  for details of the technique.
**
*/

#define G_EXTERN extern

#include "defs.h"

#define NO_PARA_DIR_ERROR -16

/*
**  Search parameters defined in .okapi_rc
**
**         P_UNIT =    4    ..... Try 6 (SW 29-May-97)
**         P_STEP =    2
**     PASSAGE_OP = "bm250"
**     DEFAULT_K1 =    1.6
**  DEFAULT_BM25B =    0.7
**      BM_TARGET = 1000
**
*/

G_EXTERN void current_time(char *);
G_EXTERN void hf_header(int, char *, int);
G_EXTERN void terminate_run(char *, char *, int);
G_EXTERN void show_time(char *, int);
G_EXTERN void malloc_error(char *, char *, char *, int);
G_EXTERN void malloc_success(char *, char *, char *, int);

void bss_search(int *np, double *max_weight) {
	int debugging = atoi(getenv("BSS_SEARCH_DEBUG"));

	int return_code;

	int r;

	int p_unit = atoi(getenv("P_UNIT"));
	int p_step = atoi(getenv("P_STEP"));
	double k1 = atof(getenv("K1"));
	double bm25b = atof(getenv("BM25B"));
	int bm_target = atoi(getenv("BM_TARGET"));

	char *bm250_and;

	int buffer_size;
	int chars_read;
	int length;

	int next_set_rec;

  int combine_term = 0;
  int nmaxwt;
  int ngw;
  int nmpw;

  int nop;

  double mpw;

  int next_set;
  int next_term = 0;
  int no_terms = 0;
  int set_count = 0;
  int temp_set;
  int top_set;

  int passage_set[MAX_QUERY_SIZE];

  char *fbuf;
  char *rbuf;
  char elapsed[RUNTIME_LENGTH];
  char ttype;
  
  char file_buffer[SMALL_BUFFER];

	char bss_command[10 * SMALL_BUFFER];
	char bss_result[SMALL_BUFFER];
  char passage_command[SMALL_BUFFER];
  
  struct search_terms *search_term;

	int hitlist_size = atoi(getenv("MAX_RECS_TO_SHOW"));
	int last_record;

  /*
  **  Current query is written to history file in:
  **
  **    gui.cc : db_search()
  **
  **  which calls the function
  **
  **    utilities.c : write_terms_to_history_file()
  **
  **  ITERATION_NO is incremented in db_search() before
  **  anything else is done.
  */
  
  show_time("bss_search", START);
  
	buffer_size = (MAX_QUERY_SIZE + 1) * sizeof(struct search_terms);

	if ((search_term = (struct search_terms *) malloc(buffer_size)) == NULL) {
		malloc_error("search.c", "bss_search", "search_tern", buffer_size);
	}

	/*
	**  First set up bm250_and operator from environment variable BM250_AND
	*/

	buffer_size = strlen(getenv("BM250_AND")) + 1;

	if ((bm250_and = (char *) malloc(buffer_size)) == NULL) {
		malloc_error("search.c", "bss_search",
								 "bm250_and", buffer_size);
	}

	sprintf(bm250_and, "%s", getenv("BM250_AND"));

  /*
  **  Read the working_query_file into struct search_term
  **  and write to current_iteration_file
  */
  
  wqf = fopen(working_query_file, "r");
  cif = fopen(current_iteration_file, "w");

  while (!feof(wqf)) {
    if (fgets(file_buffer, SMALL_BUFFER, wqf) == NULL) {
      break;
    }
    else {
      file_buffer[strlen(file_buffer) - 1] = '\0';

			if (debugging) {
				fprintf(stderr, "bss_search():: file_buffer = [%s]\n",
								file_buffer);
			}

      fbuf = file_buffer;

      sscanf(fbuf, "%c:%d:%d:%lf:%c",
						 &ttype, &(search_term + no_terms)->bss_set, &nop,
						 &(search_term + no_terms)->weight, 
						 &(search_term + no_terms)->operation);

      if (debugging) {
				fprintf(stderr,
								"bss_search():: ttype = [%c]\n",
								ttype);
				fprintf(stderr,
								"bss_search():: (search_term + %d)->bss_set = [%d]\n",
								no_terms, (search_term + no_terms)->bss_set);
				fprintf(stderr, "bss_search():: nop = [%d]\n",
								nop);
				fprintf(stderr,
								"bss_search():: (search_term + %d)->weight = [%.3f]\n",
								no_terms, (search_term + no_terms)->weight);
				fprintf(stderr,
								"bss_search():: (search_term + %d)->operation = [%c]\n",
								no_terms, (search_term + no_terms)->operation);
			}

      fprintf(cif, "%d:%f:%c\n",
							(search_term + no_terms)->bss_set,
							(search_term + no_terms)->weight,
							(search_term + no_terms)->operation);
			
      no_terms++;
    }
  }
	
  fclose(wqf);
  fclose(cif);
	
  /*
  **  Set all 'passage_set' items to NONE_ASSIGNED
  */
	for (next_set = 0; next_set < MAX_QUERY_SIZE; next_set++) {
    passage_set[next_set] = NONE_ASSIGNED;
  }
  
  sprintf(bss_command, "find");
  
  /*
  **  First perform bm25 search with target 1000
  */
  
  for (next_term = 0; next_term < no_terms; next_term++) {
    sprintf(bss_command + strlen(bss_command),
						" s=%d w=%f",
						(search_term + next_term)->bss_set,
						(search_term + next_term)->weight);
    
    /*
    **  Write an entry in the history file for each termset
    **  First update ELAPSED_TIME
    */
    
		current_time(elapsed);
    hf_header(COMMAND_NO, "search", PARAMS);
		
    fprintf(hf, ":%d:%d:%f:%c\n", 
						ITERATION_NO,
						(search_term + next_term)->bss_set,
						(search_term + next_term)->weight,
						(search_term + next_term)->operation);
		
    fflush(hf);
  }
	
  if (strcmp(bss_command, "find") != 0) {
    /*
    **  Perform bm25 search with target = bm_target (1000)
    */
    
    sprintf(bss_command + strlen(bss_command),
						" op=%s target=%d", getenv("DEFAULT_OP"), bm_target);

		if (debugging) {
			fprintf(stderr, "bss_search() :: bss_command = [%s]\n", bss_command);
		}
    
    if ((return_code = i0(bss_command, bss_result)) != 0) {
      terminate_run("bss_search()", bss_command, return_code);
    }
    
    rbuf = bss_result;
    sscanf(rbuf, "S%d np=%d", &temp_set, np);
    
		/*
		**  Properly reduce set to size bm_target
		*/
    
		sprintf(bss_command, "find s=%d top=%d", temp_set, bm_target);
    
		if ((return_code = i0(bss_command, bss_result)) != 0) {
			terminate_run("bss_search()", bss_command, return_code);
		}
		
		rbuf = bss_result;

		sscanf(rbuf, "S%d np=%d maxwt=%lf nmaxwt=%d ngw=%d mpw=%lf nmpw=%d",
					 &top_set, np, max_weight, &nmaxwt, &ngw, &mpw, &nmpw);

		/*
		sscanf(rbuf, "S%d", &top_set);
		*/

		/*
		**  'temp_set' is the original bm25 set. 
		**  'top_set' now holds the set number of the bm set 
		**  reduced in size to bm_target. 
		**  Delete 'temp_set' which is no longer needed
		*/
		
		sprintf(bss_command, "delete %d", temp_set);
		
		if ((return_code = i0(bss_command, bss_result)) != 0) {
			terminate_run("bss_search()", bss_command, return_code);
		}
    
		/*
		**  for text databases do a passage search
		**
		**  Set SHOW_SET to top_set in case there is no paragraph
		**  information for a text database.
		*/

		SHOW_SET = top_set;

		if (params_file.para_file) {
			/*
			**  There is paragraph info
			**
			**  Reset combine_term to zero.
			**
			**  'and2' each termset()->bss_set with top_set to
			**  form passage_set[]. These sets are stored in an array
			**  so that they can be deleted before leaving the program.
			**
			**  Build up passage_command as:
			**
			**  find s=passage_set[set_count] w=(termset + next_term)->weight ...
			*/
    
			sprintf(passage_command, "find");
    
			set_count = 0;
			combine_term = 0;
    
			for (next_term = 0; next_term < no_terms; next_term++) {
				sprintf(bss_command,
								"find s=%d s=%d op=%s",
								(search_term + next_term)->bss_set, top_set, bm250_and);
			
				if ((return_code = i0(bss_command, bss_result)) != 0) {
					terminate_run("bss_search()", bss_command, return_code);
				}
      
				rbuf = bss_result;
			
				sscanf(rbuf, "S%d", &passage_set[set_count]);
      
				sprintf(passage_command + strlen(passage_command),
								" s=%d w=%f",
								passage_set[set_count],
								(search_term + next_term)->weight);
			
				set_count++;
			}
    
			/*
			**  Now add extra parameters to passage_command:
			**
			**     k1:  1.2 <= k1 <= 2.0
			**
			**  bm25b:  0.6 <= bm25b <= 0.75
			**
			**  e.g. k1 = 1.6,  bm25b = 0.7
			*/
			
			sprintf(passage_command + strlen(passage_command),
							" p_unit=%d p_step=%d op=%s k1=%f bm25b=%f",
							p_unit, p_step, getenv("PASSAGE_OP"), k1, bm25b);
		
			if ((return_code = i0(passage_command, bss_result)) != 0) {
				if (return_code != NO_PARA_DIR_ERROR) {
					terminate_run("bss_search()", passage_command, return_code);
				}
			}
    
			if (return_code != NO_PARA_DIR_ERROR) {
				rbuf = bss_result;
				
				sscanf(rbuf, "S%d np=%d maxwt=%lf nmaxwt=%d ngw=%d mpw=%lf nmpw=%d",
							 &temp_set, np, max_weight, &nmaxwt, &ngw, &mpw, &nmpw);
				
				if (*max_weight > mpw) {
					*max_weight = mpw;
				}
				
				if (debugging) {
					fprintf(stderr, "bss_search() : *max_weight = %.3f\n", *max_weight);
				}
				
				SHOW_SET = temp_set;
				
				/*
				**  Delete 'top_set' and 'passage_sets'
				*/
				
				sprintf(bss_command, "delete %d", top_set);
				
				if ((return_code = i0(bss_command, bss_result)) != 0) {
					terminate_run("bss_search()", bss_command, return_code);
				}
				
				for (next_set = 0; next_set < MAX_QUERY_SIZE; next_set++) {
					if (passage_set[next_set] != NONE_ASSIGNED) {
						sprintf(bss_command, "delete %d", passage_set[next_set]);
						
						if ((return_code = i0(bss_command, bss_result)) != 0) {
							terminate_run("bss_search()", bss_command, return_code);
						}
					}
				}
			}
		}

		if (debugging) {
			last_record = (*np > hitlist_size) ? hitlist_size : *np;
			
			for (r = 0; r < last_record; r++) {
				sprintf(bss_command, "show f=0 r=%d", r);
				
				if ((return_code = i0(bss_command, bss_result)) != 0) {
					terminate_run("bss_search()", bss_command, return_code);
				}
				
				fprintf(stderr, "%d) %s\n", r, bss_result);
			}
		}

    /*
    **  Complete history file entries
    **  First update ELAPSED_TIME
    */

		current_time(elapsed);

    hf_header(COMMAND_NO, "docset", PARAMS);

    fprintf(hf, ":%d:%d:%d:%f:%d:%d:%f:%d\n",
						ITERATION_NO, SHOW_SET,
						*np, *max_weight, nmaxwt, ngw, mpw, nmpw);
    fflush(hf);
    
    SHOW_SET_RECNO = 0;
    NEW_TERMS = FALSE;
    
    /*
		**  Messages for display in tcl command window
		*/
  }
  else {
    printf("bss_search: There are no terms in the termset\n");
  }
 
	free(search_term);
	free(bm250_and);

  show_time("bss_search", FINISH);
  
  NEW_ITERATION = FALSE;
}
