/*
**  show.c
**
**  fsf : final_show_file      file to be shown in text window
**
**  Files for term extraction. Initially one term per record is
**  written to theses files. Some of these may be phrases (incl.
**  GSL phrases) which are eventually removed from the file since
**  there is no way of picking these out of a record by highlight
**  alone => need to be reweighted separately (NOTE: this can lead
**  to incorrect (r, wgt, rsv) values for documents that are judged
**  relevant on a passage only since phrases may exist in the full
**  record but not in the passage.
*/

#define G_EXTERN extern

#include "defs.h"


#undef MAX_DISPLAY_LENGTH
#define MAX_DISPLAY_LENGTH 77

#undef MAX_WORD_LENGTH
#define MAX_WORD_LENGTH 80

G_EXTERN void construct_doclength_field(char *, int, int, int, int);

G_EXTERN void show_time(char *, int);
G_EXTERN void determine_doc_length(int, int, int *, int *, int*);
G_EXTERN void terminate_run(char *, char *, int);
G_EXTERN void parse_header(char *, char *, 
													 struct header *,
													 struct dir_records *,
													 struct passage_records *,
													 struct highlight_records *,
													 int *);

G_EXTERN void screen_error();
G_EXTERN void history_file_error(char *);
G_EXTERN void convert_date(char *, char *);
G_EXTERN void hf_header(int, char *, int);
G_EXTERN void malloc_error(char *, char *, char *, int);
G_EXTERN void malloc_success(char *, char *, char *, int);

void show_record(int show_set, int show_rec, double max_weight,
								 int *doc_part, char final_show_file[],
								 int *full_doclen, int *part_doclen) {

	int debug = atoi(getenv("SHOW_DEBUG"));

	int return_code;

	int buffer_size;
	int max_doclength;

  int char_count = 0;
  int chars_read;
  int doc_length;
	int passage_length;
	int page_size = atoi(getenv("CHARS_PER_PAGE"));

  int doc_start;
  int doc_end;
  double doc_weight;
  int hi_on = FALSE;
  int length;
  int next_term = 0;

	int field_no;
	int field_pos;

	int header_length;

	int no_blanks;

  int pass_on = FALSE;
  int sub_passage = TRUE;
  int part_only = FALSE;
  int pre_chars = EXTRA_CHARS;
  int post_chars = EXTRA_CHARS;
  int total_chars = 0;

  int display_doclen;
  int display_paslen;

	int FULLDOC;
	int PASSAGE;

	int field_end;
	int nf;

	/*
	**  variables used when determining the length of the 'doclength_field'
	*/

	int plen;
	int dlen;

	char normalised_weight[6];
  char extracted_word[MAX_WORD_LENGTH];

  char *doc_ptr;

	char *show_line;
  char *docid;
	char *creation_date;
	char *doclength_field;

  char *fulldoc_file;
  char *passage_file;
	char *full_minus_pass_file;

  char *blanks;
  char *divider;

	char *bss_command;
	char *bss_result;

	char *returned_doc;

  /*
  **  Show header data
  */

  struct header *show_head;

	struct dir_records *doc_fields;
	struct passage_records *passage;
	struct highlight_records *hl_terms;

  FILE *sf;
  FILE *fd;
  FILE *pd;
	FILE *fmp;
  FILE *fsf;

	/*
	**  DOC_THRESHOLD set in ".okapi_rc"
	*/

	max_doclength = atoi(getenv("DOC_THRESHOLD"));

	buffer_size = SMALL_BUFFER;

	if ((bss_command = (char *) malloc(buffer_size)) == NULL) {
		malloc_error("show.c", "show_record",
								 "bss_command", buffer_size);
	}
	else {
		if (debug) {
			malloc_success("show.c", "show_record",
									 "bss_command", buffer_size);
		}
	}

	buffer_size = MAX_DISPLAY_LENGTH + 1;

	if ((blanks = (char *) malloc(buffer_size)) == NULL) {
		malloc_error("show.c", "show_record",
								 "blanks", buffer_size);
	}
	else {
		if (debug) {
			malloc_success("show.c", "show_record",
									 "blanks", buffer_size);
		}
	}

	buffer_size = MAX_DISPLAY_LENGTH + 1;

	if ((divider = (char *) malloc(buffer_size)) == NULL) {
		malloc_error("show.c", "show_record",
								 "divider", buffer_size);
	}
	else {
		if (debug) {
			malloc_success("show.c", "show_record",
									 "divider", buffer_size);
		}
	}

  memset(blanks, ' ', MAX_DISPLAY_LENGTH);
  *(blanks + MAX_DISPLAY_LENGTH) = '\0';

	memset(divider, '_', MAX_DISPLAY_LENGTH);
	*(divider + MAX_DISPLAY_LENGTH) = '\0';

  /*
  **  'show_file' and 'fullrecs_file' already include the pid
  */
	
	buffer_size = strlen(show_file) + 6;

	if ((fulldoc_file = (char *) malloc(buffer_size)) == NULL) {
		malloc_error("show.c", "show_record",
								 "fulldoc_file", buffer_size);
	}
	else {
		if (debug) {
			malloc_success("show.c", "show_record",
									 "fulldoc_file", buffer_size);
		}
	}

	if ((passage_file = (char *) malloc(buffer_size)) == NULL) {
		malloc_error("show.c", "show_record",
								 "passage_file", buffer_size);
	}
	else {
		if (debug) {
			malloc_success("show.c", "show_record",
									 "passage_file", buffer_size);
		}
	}

	if ((full_minus_pass_file = (char *) malloc(buffer_size)) == NULL) {
		malloc_error("show.c", "show_record",
								 "full_minus_pass_file", buffer_size);
	}
	else {
		if (debug) {
			malloc_success("show.c", "show_record",
									 "full_minus_pass_file", buffer_size);
		}
	}

  sprintf(fulldoc_file, "%s.full", show_file);
  sprintf(passage_file, "%s.pass", show_file);
  sprintf(full_minus_pass_file, "%s.fminusp", show_file);

  sprintf(final_show_file, "%s", fullrecs_file);

  show_time("show_record", START);

	buffer_size = sizeof(struct header);

	if ((show_head = (struct header *) malloc(buffer_size)) == NULL) {
		malloc_error("show.c", "show_record",
								 "show_head", buffer_size);
	}
	else {
		if (debug) {
			malloc_success("show.c", "show_record",
									 "show_head", buffer_size);
		}
	}

	buffer_size = (params_file.nf + 1) * sizeof(struct dir_records);

	if ((doc_fields  =	(struct dir_records *) malloc(buffer_size)) == NULL) {
		malloc_error("show.c", "show_record",
								 "doc_fields", buffer_size);
	}
	else {
		if (debug) {
			malloc_success("show.c", "show_record",
									 "doc_fields", buffer_size);
		}
	}

	buffer_size = MAX_PASSAGES * sizeof(struct passage_records);

	if ((passage = (struct passage_records *) malloc(buffer_size)) == NULL) {
		malloc_error("show.c", "show_record",
								 "passage", buffer_size);
	}
	else {
		if (debug) {
			malloc_success("show.c", "show_record",
									 "passage", buffer_size);
		}
	}

	buffer_size = MAX_HIGHLIGHT_TERMS * sizeof(struct highlight_records);

	if ((hl_terms = (struct highlight_records *) malloc(buffer_size)) == NULL) {
		malloc_error("show.c", "show_record",
								 "hl_terms", buffer_size);
	}
	else {
		if (debug) {
			malloc_success("show.c", "show_record",
									 "hl_terms", buffer_size);
		}
	}

  /*
  **  First determine doc & passage lengths. Although don't need
	**  passage_length to determine results buffer size the function
	**  used is the same as that in hitlist.c where it is needed to
	**  determine whether or not to include the doc in the hitlist.
  */
	
	determine_doc_length(show_set, show_rec,
											 &header_length, &doc_length, &passage_length);

	if (debug) {
		fprintf(stderr, "show_record() ::  header_length = %d\n", header_length);
		fprintf(stderr, "show_record() ::     doc_length = %d\n", doc_length);
		fprintf(stderr, "show_record() :: passage_length = %d\n", passage_length);
	}

	buffer_size = header_length + doc_length + 1;

	if ((bss_result = (char *) malloc(buffer_size)) == NULL) {
		malloc_error("show.c", "show_record",
								 "bss_result", buffer_size);
	}
	else {
		if (debug) {
			malloc_success("show.c", "show_record",
									 "bss_result", buffer_size);
		}
	}

	buffer_size = doc_length + 1;

	if ((returned_doc = (char *) malloc(buffer_size)) == NULL) {
		malloc_error("show.c", "show_record",
								 "returned_doc", buffer_size);
	}
	else {
		if (debug) {
			malloc_success("show.c", "show_record",
									 "returned_doc", buffer_size);
		}
	}

  /*
  **  First show record
  */
	
  sprintf(bss_command,
					"show s=%d f=%d r=%d",
					show_set, atoi(getenv("PASSAGE_SHOW_FORMAT")), show_rec);
	
  if ((return_code = i0(bss_command, bss_result)) < 0) {
    terminate_run("show_record", bss_command, return_code);
  }
	
  /*
  **  Now parse header information using the function 'parse_header()'
  **  in 'utilities.c'
	**
  **  'doc_ptr' will contain the whole document, excluding the
  **  header information, i.e. docid onwards.
  **  It will be written to the file 'final_show_file'
  */

	if (debug) {
		fprintf(stderr, "show_record() ::  bss_result = [%s]\n", bss_result);
	}

  parse_header(bss_result, returned_doc, show_head,
							 doc_fields, passage, hl_terms, &next_term);

	if (debug) {
		fprintf(stderr, "show_record() ::  returned_doc = [%s]\n", returned_doc);
	}

	*full_doclen = doc_length;

	if (show_head->no_passage_recs == 2) {
		PASSAGE =
			((passage + 0)->length > (passage + 1)->length) ? 1 : 0;
		
		FULLDOC = (PASSAGE == 0) ? 1 : 0;
		
		*part_doclen = (passage + PASSAGE)->length;
		part_only = (doc_length >= max_doclength &&
								 passage_length < max_doclength);
	}
	else {
		*part_doclen = doc_length;
		part_only = FALSE;
	}

	if (!part_only) {
		next_term = 0;
	}

	doc_ptr = returned_doc;

  /*
  **  next_term will store the number of the first highlighted
  **  term in the record. It is always 0 when the full record
  **  is shown, and set to the first highlight within the passage
  **  of doclen > max_doclength
  **
  **  -----------------------------------------------------------
  **
  **  Set boolean variable 'sub_passage' according
  **  to whether there is a sub-passage or not.
  */

  sub_passage = (show_head->no_passage_recs == 2);

  /*
  **  Calculate document and passage lengths in pages
	**  determined by CHARS_PER_PAGE (set in .okapi_rc and read
	**  into the int variable page_size - probably 2000 chars)
  */

  
	if (doc_length <= page_size) {
		display_doclen = 1;
	}
	else {
		display_doclen
			= (int) floor(((float) doc_length / (float) page_size) + 0.5);
  }

	if (show_head->no_passage_recs == 2) {
		if (passage_length <= page_size) {
			display_paslen = 1;
		}
		else {
			display_paslen
				= (int) floor(((float) passage_length / (float) page_size) + 0.5);
		}
	}

  /*
  **  Set 'doc_part' to:
  **
  **  NONE_ASSIGNED (default)
  **  0 -- passage and fullrec are the same length
  **  1 -- part_only (full doc too long but passage ok)
	**
	**  display_doclen & display_paslen are both calculated to the
	**  nearest page.
  */

	*doc_part = NONE_ASSIGNED;

  if ((doc_length == passage_length) ||
			(display_doclen == display_paslen)) {
		/*
		**  Full doc & passage are approximately the same length
		*/
		*doc_part = 0;
	}
	else if (part_only) {
		/*
		**  Full doc too long but passage ok
		*/
		*doc_part = 1;
	}

  /*
  **  Open 'final_show_file' - the file that will be loaded
  **  into the fullrec text widget.
  */

  if ((fsf = fopen(final_show_file, "w")) == (FILE *)0) {
    sprintf(ERROR_MESSAGE, "Can't open %s for writing", final_show_file);
    screen_error();
    history_file_error("show");
    exit(-1);
  }

  total_chars = 0;

	/*
	**  Pick out docid from field 1
	*/

	buffer_size = (doc_fields + 0)->length + 1;

	if ((docid = (char *) malloc(buffer_size)) == NULL) {
		malloc_error("show.c", "show_record",
								 "docid", buffer_size);
	}
	else {
		if (debug) {
			malloc_success("show.c", "show_record",
									 "docid", buffer_size);
		}
	}

	memcpy(docid, doc_ptr, (doc_fields + 0)->length);

	/*
	**  "ai" type databases don't contain newline chars.
	**  Other types contain newline chars at the end of each
	**  displayed line. Hence the docid string needs the
	**  newline character to be removed.
	*/
	
	if (strncmp(params_file.db_type, "ai", 2) == 0) {
		*(docid + (doc_fields + 0)->length) = '\0';
	}
	else {
		*(docid + (doc_fields + 0)->length - 1) = '\0';
	}
	
	/*
	**  Move past docid (field 0)
	*/

	doc_ptr += (doc_fields + 0)->length;
	total_chars += (doc_fields + 0)->length;

  /*
  **  Build up header line to document
  **
  **  The function MapWeight() maps the weight range
  **  (minwt..max_weight) onto the range (1..1000)
  */

	sprintf(normalised_weight, "%d",
					MapWeight(show_head->weight, max_weight));

	/*
	**  Now make up doclength field
	**
	**  First determine number of spaces for numbers,
	**  including an 's' if number > 1
	**
	**  Extra characters are:
	**
	**      zero length passage: 'Length:  page', = 13 + 1 = 14
	**  non-zero length passage: 'Length:  page (~  page)', = 23 + 1 = 24
	*/

	buffer_size = MAX_DOCID_LENGTH + 1;

	if ((doclength_field = (char *) malloc(buffer_size)) == NULL) {
		malloc_error("show.c", "show_record",
								 "doclength_field", buffer_size);
	}
	else {
		if (debug) {
			malloc_success("show.c", "show_record",
									 "doclength_field", buffer_size);
		}
	}
	
	if (show_head->no_passage_recs == 2) {
		construct_doclength_field(doclength_field,
															show_head->no_passage_recs,
															(passage + PASSAGE)->length,
															(passage + FULLDOC)->length,
															page_size);
	}
	else {
		construct_doclength_field(doclength_field,
															show_head->no_passage_recs,
															doc_length,
															doc_length,
															page_size);
	}
		
	/*
	**  Allocate storage for show_line
	*/

	buffer_size = MAX_DISPLAY_LENGTH + 1;

	if ((show_line = (char *) malloc(buffer_size)) == NULL) {
		malloc_error("show.c", "show_record",
								 "show_line", buffer_size);
	}
	else {
		if (debug) {
			malloc_success("show.c", "show_record",
									 "show_line", buffer_size);
		}
	}

	sprintf(show_line,
					"Record %d          Weight %s",
					show_head->set_recno + 1, normalised_weight);
	
  no_blanks =
		MAX_DISPLAY_LENGTH - strlen(show_line) - strlen(doclength_field);
	
	strncat(show_line, blanks, no_blanks);
	sprintf(show_line + strlen(show_line), "%s", doclength_field);
	
	fprintf(fsf, "%s\n", show_line);

	free(doclength_field);
	free(show_line);

	fprintf(fsf, "%s\n", divider);

  doc_weight = show_head->weight;
	
  hf_header(COMMAND_NO++, "show", PARAMS);

  /*
  **  Since ITERATION_NO has already been incremented we
  **  must write ITERATION_NO - 1 to history file.
  */

	if (show_head->no_passage_recs == 2) {
		fprintf(hf, ":%d:%d:%s:%.3f:%d",
						ITERATION_NO, show_rec, docid,
						(passage + FULLDOC)->weight, (passage + FULLDOC)->length);
	}
	else {
		fprintf(hf, ":%d:%d:%s:%.3f:%d",
						ITERATION_NO, show_rec, docid,
						show_head->weight, doc_length);
	}

  fflush(hf);
	
	/*
	**  Extra bit to add in date field
	*/

  if (part_only) {
    /*
    **  If document > max_doclength only show the best passage
    */
    if (total_chars == (passage + PASSAGE)->offset) {
      pre_chars = 0;
    }
    else if (total_chars + pre_chars > (passage + PASSAGE)->offset) {
      pre_chars = (passage + PASSAGE)->offset - total_chars;
    }
		
    if (((passage + FULLDOC)->length -
				 (passage + PASSAGE)->length -
				 (passage + PASSAGE)->offset) < post_chars) {
      post_chars =
				(passage + FULLDOC)->length -
				(passage + PASSAGE)->length -
				(passage + PASSAGE)->offset;
    }

    doc_ptr += (passage + PASSAGE)->offset - total_chars - pre_chars;
    total_chars += (passage + PASSAGE)->offset - total_chars - pre_chars;
    length = (passage + PASSAGE)->length + post_chars + pre_chars;  
    doc_start = (passage + PASSAGE)->offset;
    doc_end = (passage + PASSAGE)->offset + (passage + PASSAGE)->length;
  }
  else {
    /*
    **  Show entire document
    */
		
		pre_chars = (doc_fields + 1)->length;
    doc_start = (doc_fields + 0)->length;
    doc_end = doc_length;
    length = doc_length - (doc_fields + 0)->length;
    post_chars = 0;
  }

  /*
  **  Move to start of text to display
  */

  hi_on = FALSE;
  pass_on = FALSE;
  char_count = 0;
	
  if (debug ) {
		printf("doc_ptr = [%s]\n", doc_ptr);
		printf("total_chars = [%d], hi_on [%d], pass_on [%d], next_term = [%d]\n",
					 total_chars, hi_on, pass_on, next_term);
	}

  if ((fd = fopen(fulldoc_file, "w")) == (FILE *)0) {
    sprintf(ERROR_MESSAGE, "Can't open %s for writing", fulldoc_file);
    screen_error();
    history_file_error("show");
    exit(-1);
  }
	
  if ((pd = fopen(passage_file, "w")) == (FILE *)0) {
    sprintf(ERROR_MESSAGE, "Can't open %s for writing", passage_file);
    screen_error();
    history_file_error("show");
    exit(-1);
  }
	
  if ((fmp = fopen(full_minus_pass_file, "w")) == (FILE *)0) {
    sprintf(ERROR_MESSAGE, "Can't open %s for writing", full_minus_pass_file);
    screen_error();
    history_file_error("show");
    exit(-1);
  }
	
  if (part_only) {
    fprintf(fsf, " .....\n");

		if (debug) {
			printf(" .....");
		}
  }

  /*
  **  Now write the <TEXT> field to the 'final_show_file',
	**  inserting passage and highlight on/off codes as appropriate.
  */

	for ( ; ; ) {
    /*
    **  First test to see if it is the start of a sub passage.
    **  If not, and we are past the start of the best passage
    **  and it is the end of the line, put a <Q> at the end.
    */
		
		for (field_no = 0; field_no < show_head->no_dir_recs; field_no++) {
			field_pos = (doc_fields + field_no)->offset +
				(doc_fields + field_no)->length;

			if (total_chars == field_pos) {
				if (field_no > 0) {
					fprintf(fsf, "\n\n");
					/*
					if (strncmp(params_file.db_type, "ai", 2) == 0) {
						fprintf(fsf, "\n");
					}
					*/
				}

				if (field_no < (show_head->no_dir_recs - 1)) {
					fprintf(fsf, "%d: ", field_no + 1);
				}
			}

		}
    if (sub_passage) {
      if (((passage + PASSAGE)->offset == total_chars) &&
					(passage_length < doc_length)) {
				fprintf(fsf, "<P>");
				pass_on = TRUE;

				if (debug) {
					printf("<P>");
				}
      }
    }
		
    /*
    **  Check for start of highlighted term
    */
    
    if (next_term < show_head->no_highlight_recs) {
      if ((hl_terms + next_term)->length > 0) {
				if ((hl_terms + next_term)->offset == total_chars) {
					hi_on = TRUE;

					if (pass_on) {
						fprintf(fsf, "<Q>");

						if (debug) {
							printf("<Q>");
						}
					}
					
					fprintf(fsf, "<S>");

					if (debug) {
						printf("<S>");
					}
				}
      }
    }
    
    /*
    **  Write the character to 'final_show_file',
    **  'passage_file' and 'fulldoc_file' accordingly.
    **  First check if it's a <nl> char within a best passage.
    */
    
    if ((pass_on) && (*doc_ptr == '\n')) {
      /*
      **  <nl> in the within a passage - write a space
      **  to both fd & pd.
      */
      if (hi_on) {
				/*
				**  Terms already preceded by a <Q><S>
				*/
				fprintf(fsf, "<E>\n<S>");
				fprintf(fd, " ");
				fprintf(pd, " ");
				char_count++;

				if (debug) {
					printf("<E>\n<S>");
				}
      }
      else {
				fprintf(fsf, "<Q>\n<P>");
				fprintf(fd, "\n");
				fprintf(pd, "\n");
				char_count = 0;

				if (debug) {
					printf("<Q>\n<P>");
				}
      }
    }
    else if ((!pass_on) && (*doc_ptr == '\n')) {
      /*
      **  <nl> outside of a passage - write to pass_full only
      */
      if (hi_on) {
				fprintf(fsf, "<E>\n<S>");
				fprintf(fd, " ");
				fprintf(fmp, " ");
				char_count++;

				if (debug) {
					printf("<E>\n<S>");
				}
      }
      else {
				fprintf(fsf, "%c", *doc_ptr);
				fprintf(fd, "%c", *doc_ptr);
				fprintf(fmp, "%c", *doc_ptr);
				char_count = 0;

				if (debug) {
					printf("%c", *doc_ptr);
				}
      }
    }
    else {
      /*
      **  Any other character
      */
			
      if (isprint(*doc_ptr)) {
				fprintf(fsf, "%c", *doc_ptr);

				if (debug) {
					printf("%c", *doc_ptr);
				}
      }
      
      if (*doc_ptr == '-') {
				/*
				**  Test for hyphenated word
				*/
				if (char_count < 3) {
					/*
					**  Write to fd
					*/
					fprintf(fd, "%c", *doc_ptr);
					if (pass_on) {
						/*
						**  Write to pd as well
						*/
						fprintf(pd, "%c", *doc_ptr);
					}
					else {
						fprintf(fmp, "%c", *doc_ptr);
					}
					char_count++;
				}
				else {
					/*
					**  Write to fd
					*/
					fprintf(fd, "\n");
					if (pass_on) {
						/*
						**  Write to pd as well
						*/
						fprintf(pd, "\n");
					}
					else {
						fprintf(fmp, "\n");
					}
					char_count = 0;
				}
      }
      else if ((*doc_ptr == ' ') && (!hi_on)) {
				/*
				**  Write to fd
				*/
				fprintf(fd, "\n");
				if (pass_on) {
					/*
					**  Write to pd as well
					*/
					fprintf(pd, "\n");
				}
				else {
					fprintf(fmp, "\n");
				}
				char_count = 0;
      }
      else {
				/*
				**  Write to fd
				*/
				if (isprint(*doc_ptr)) {
					fprintf(fd, "%c", *doc_ptr);
					if (pass_on) {
						/*
						**  Write to pd as well
						*/
						fprintf(pd, "%c", *doc_ptr);
					}
					else {
						fprintf(fmp, "%c", *doc_ptr);
					}
					char_count++;
				}
      }
    }
		
    /*
    **  Increment 'total_chars' and move past character.
    */
    
    total_chars++;
    doc_ptr++;

    /*
    **  Check for end of highlighted term
    */
    
    if (next_term < show_head->no_highlight_recs) {
			if (((hl_terms + next_term)->offset + (hl_terms + next_term)->length)
					== total_chars) {
				fprintf(fsf, "<E>");
				hi_on = FALSE;
				
				if (debug) {
					printf("<E>");
				}
				
				if (pass_on) {
					fprintf(fsf, "<P>");
					
					if (debug) {
						printf("<P>");
					}
				}
				
				next_term++;
			}
		}
    
    /*
    **  If the end of the best passage print a <Q>
    */

    if (sub_passage) {
      if ((passage + PASSAGE)->offset +	(passage + PASSAGE)->length
					== total_chars + 1) {
				fprintf(fsf, "<Q>");
				pass_on = FALSE;

				if (debug) {
					printf("<Q>");
				}
      }
    }

		if (*doc_ptr == '\0') {
			fprintf(fsf, "%c", *doc_ptr);

			if (debug) {
				printf("%c", *doc_ptr);
			}

			break;
		}

		/*
		**  For "ai" type databases print an extra linefeed at 
		**  the end of a field.
		*/

		if (strcmp(params_file.db_type, "ai") == 0) {
			field_end = 0;
			
			for (nf = 0; nf < show_head->no_dir_recs; nf++) {
				field_end += (doc_fields + nf)->length;
				
				if (total_chars == field_end) {
					fprintf(fsf, "\n\n");
				}
			}
		}
  }

  if (pass_on) {
    fprintf(fsf, "<Q>");
    pass_on = FALSE;

		if (debug) {
			printf("<Q>");
		}
  }
  
  if (part_only) {
    fprintf(fsf, "\n.....");

		if (debug) {
			printf("\n.....>");
		}
  }

  fclose(fsf);
  
  fprintf(fd, "\n");
  fclose(fd);
  
  fprintf(pd, "\n");
  fclose(pd);

  fprintf(fmp, "\n");
  fclose(fmp);

  /*
  **  Remove phrases from 'fulldoc_file'. Write temporarily to
  **  'show_file' which is finally renamed to 'fulldoc_file'
  */
  
  if ((fd = fopen(fulldoc_file, "r")) == (FILE *)0) {
    sprintf(ERROR_MESSAGE, "Can't open %s for reading", fulldoc_file);
    screen_error();
    history_file_error("show");
    exit(-1);
  }
  
  if ((sf = fopen(show_file, "w")) == (FILE *)0) {
    sprintf(ERROR_MESSAGE, 
						"Can't open show_file (%s) for writing", show_file);
    screen_error();
    history_file_error("show");
    exit(-1);
  }
  
  while (!feof(fd)) {
    if (fgets(extracted_word, TITLE_CHARS, fd) == NULL) {
      break;
    }
    
    extracted_word[strlen(extracted_word) - 1] = '\0';
    
    if (strlen(extracted_word) > 0) {
			fprintf(sf, "%s\n", extracted_word);
    }
  }
  
  fclose(fd);
  fclose(sf);
  
  /*
  **  Delete "fulldoc_file".
  **  Rename "show_file" -> "fulldoc_file"
  */
  
  sprintf(OS_COMMAND, "%s %s", RM, fulldoc_file);

  if (system(OS_COMMAND) < 0) {
    perror(OS_COMMAND);
  }
  
  sprintf(OS_COMMAND, "%s %s %s", MV, show_file, fulldoc_file);
  if (system(OS_COMMAND) < 0) {
    perror(OS_COMMAND);
  }
  
  /*
  **  Remove phrases from 'passage_file'. Write temporarily to
  **  'show_file' which is finally renamed to 'fulldoc_file'
  */
  
  if ((pd = fopen(passage_file, "r")) == (FILE *)0) {
    sprintf(ERROR_MESSAGE, "Can't open %s for reading", passage_file);
    screen_error();
    history_file_error("show");
    exit(-1);
  }
  
  if ((sf = fopen(show_file, "w")) == (FILE *)0) {
    sprintf(ERROR_MESSAGE, 
						"Can't open show_file (%s) for writing", show_file);
    screen_error();
    history_file_error("show");
    exit(-1);
  }
  
  while (!feof(pd)) {
    if (fgets(extracted_word, TITLE_CHARS, pd) == NULL) {
      break;
    }
    
    extracted_word[strlen(extracted_word) - 1] = '\0';
    
    if (strlen(extracted_word) > 0) {
      fprintf(sf, "%s\n", extracted_word);
    }
  }
  
  fclose(pd);
  fclose(sf);
	
  /*
  **  Delete "passage_file".
  **  Rename "show_file" -> "passage_file"
  */
  
  sprintf(OS_COMMAND, "%s %s", RM, passage_file);
  if (system(OS_COMMAND) < 0) {
    perror(OS_COMMAND);
  }

  sprintf(OS_COMMAND, "%s %s %s", MV, show_file, passage_file);
  if (system(OS_COMMAND) < 0) {
    perror(OS_COMMAND);
  }
	
  /*
  **  Remove phrases from 'full_minus_pass_file'. Write temporarily to
  **  'show_file' which is finally renamed to 'fulldoc_file'
  */
  
  if ((fmp = fopen(full_minus_pass_file, "r")) == (FILE *)0) {
    sprintf(ERROR_MESSAGE, "Can't open %s for reading", full_minus_pass_file);
    screen_error();
    history_file_error("show");
    exit(-1);
  }
  
  if ((sf = fopen(show_file, "w")) == (FILE *)0) {
    sprintf(ERROR_MESSAGE, 
						"Can't open show_file (%s) for writing", show_file);
    screen_error();
    history_file_error("show");
    exit(-1);
  }
  
  while (!feof(fmp)) {
    if (fgets(extracted_word, TITLE_CHARS, fmp) == NULL) {
      break;
    }
    
    extracted_word[strlen(extracted_word) - 1] = '\0';
    
    if (strlen(extracted_word) > 0) {
			fprintf(sf, "%s\n", extracted_word);
    }
  }
  
  fclose(fmp);
  fclose(sf);
  
  /*
  **  Delete "fulldoc_file".
  **  Rename "show_file" -> "fulldoc_file"
  */
  
  sprintf(OS_COMMAND, "%s %s", RM, full_minus_pass_file);

  if (system(OS_COMMAND) < 0) {
    perror(OS_COMMAND);
  }
  
  sprintf(OS_COMMAND, "%s %s %s", MV, show_file, full_minus_pass_file);
  if (system(OS_COMMAND) < 0) {
    perror(OS_COMMAND);
  }
  
  /*
  **  Rename passage_file or fulldoc_file to show_file (i.e. the file
  **  of extracted terms). If the whole document is to be shown this
  **  defaults to the 'fulldoc_file'. If the document is longer than
  **  max_doclength then only the 'passage_file' is needed.
  */

  if (part_only) {
    sprintf(OS_COMMAND, "%s %s %s", CP, passage_file, show_file);
    if (system(OS_COMMAND) < 0) {
      perror(OS_COMMAND);
    }  
  }
  else {
    sprintf(OS_COMMAND, "%s %s %s", CP, fulldoc_file, show_file);
    if (system(OS_COMMAND) < 0) {
      perror(OS_COMMAND);
    }  
  }

	/*
	**  Return string to gui.cc::show_rec() of the form:
	**
	**  <filename>:<[F:P]>:<received_date>:<pathname>
	**
	**  <filename> is the file contating the text to be displayed.
	**  <F|P> indicates which is the highest weighted part -
	**  the (F)ull document or the (P)assage.
	**  <received_date> is the time the record was added to the database
	*/

  if (sub_passage) {
    /*
    **  There are two passages
    */
    if ((passage + PASSAGE)->weight > (passage + FULLDOC)->weight) {
      /*
      **  Sub-passage has the highest weight
      */
      sprintf(final_show_file + strlen(final_show_file), ":P");
    }
    else {
      sprintf(final_show_file + strlen(final_show_file), ":F");
    }
  }
  else {
		sprintf(final_show_file + strlen(final_show_file), ":F");
  }

	free(returned_doc);
	free(docid);
	free(bss_command);
	free(bss_result);
	free(show_head);
	free(doc_fields);
	free(passage);
	free(hl_terms);
	free(fulldoc_file);
	free(passage_file);
	free(blanks);
	free(divider);

  show_time("show_record", FINISH);
}

