/*
**  sorts.c
**
**  Quicksort functions. Moved from utilities.c
**
**  mg - Apr 1997
*/

#define G_EXTERN extern

#include "defs.h"

int compare_rsv(struct query_terms *ts1, struct query_terms *ts2) {
	/*
	**	Quicksort comparison function for termset[]. Sort keys:
	**
	**		1) rsv desc, (2) weight desc, (3) parsed asc
	*/
	
	register int to_swap = FALSE;
	double multiplier = 1000.0;
	int irsv1;
	int irsv2;
	int iwgt1;
	int iwgt2;

	irsv1 = (int) multiplier * ts1->rsv;
	irsv2 = (int) multiplier * ts2->rsv;
	iwgt1 = (int) multiplier * ts1->weight;
	iwgt2 = (int) multiplier * ts2->weight;

	if (toupper(ts1->term_type) == toupper(ts2->term_type)) {
		if (irsv1 == irsv2) {
			if (iwgt1 == iwgt2) {
				if (ts1->little_r == ts2->little_r) {
					if (ts1->np == ts2->np) {
						to_swap = (strcmp(ts1->parsed, ts2->parsed) > 0);
					}
					else {
						to_swap = (ts1->np > ts2->np);
					}
				}
				else {
					to_swap = (ts1->little_r < ts2->little_r);
				}
      }
      else {
				to_swap = (iwgt1 < iwgt2);
      }
    }
    else {
      to_swap = (irsv1 < irsv2);
    }
  }
  else {
    to_swap = (toupper(ts1->term_type) < toupper(ts2->term_type));
  }
	
  return to_swap;
}


int compare_parsed(struct query_terms *ts1, struct query_terms *ts2) {
  /*
	**  Quicksort comparison function for termset[]. Sort keys:
	**
	**    1) rsv desc, (2) weight desc, (3) parsed asc
	*/
	
  register int to_swap = FALSE;
	
  to_swap = (strcmp(ts1->parsed, ts2->parsed) > 0);
	
  return to_swap;
}


void sort_terms(struct query_terms *tset, int no_terms, int sort_flag) {
  int i;
	
  /*
  **  sort tset according to sort_flag,
  **
  **       TSORT_RSV:  rsv desc
  **    TSORT_PARSED:  paesed asc
  */
	
  if (sort_flag == TSORT_RSV) {
    (void) qsort((char *) tset, no_terms, sizeof(*tset), compare_rsv);
  }
  else if (sort_flag == TSORT_PSD) {
    (void) qsort((char *) tset, no_terms, sizeof(*tset), compare_parsed);
  }
	
  /*
  **  Reallocate term_nos according to the new order.
  **
  */
	
  for (i = 0; i < no_terms; i++) {
    (tset + i)->term_no = i;
  }
}


int compare_wqu(struct working_query *ts1, struct working_query *ts2) {
  /*
	**  Quicksort comparison function for termset[]. Sort keys:
	**
	**    1) rsv desc, (2) weight desc, (3) parsed asc
	*/
  register int to_swap = FALSE;
	double multiplier = 1000.0;
	int irsv1;
	int irsv2;
	int iwgt1;
	int iwgt2;

	irsv1 = (int) multiplier * ts1->rsv;
	irsv2 = (int) multiplier * ts2->rsv;
	iwgt1 = (int) multiplier * ts1->weight;
	iwgt2 = (int) multiplier * ts2->weight;
	
  if (irsv1 == irsv2) {
    if (iwgt1 == iwgt2) {
      if (ts1->little_r == ts2->little_r) {
				to_swap = (ts1->np > ts2->np);
      }
      else {
				to_swap = (ts1->little_r < ts2->little_r);
      }
    }
    else {
      to_swap = (iwgt1 < iwgt2);
    }
  }
  else {
    to_swap = (irsv1 < irsv2);
  }
	
  return to_swap;
}


void sort_current_query(struct working_query *current, int no_terms) {
  /*
  **  Sort the working query
  */
  (void) qsort((char *) current, no_terms, sizeof(*current), compare_wqu);
}


int compare_user_rels_wgt(struct user_rj *ur1, struct user_rj *ur2) {
  /*
  **  Quick sort Comparison function for user rels. Sort key:
  **
  **  Used to sort by relj asc then weight desc.
  **  Now only sort by weight desc
  */
	
  register int to_swap = FALSE;
	double multiplier = 1000.0;

	int iwgt1 = (int) multiplier * ur1->weight;
	int iwgt2 = (int) multiplier * ur2->weight;

	/*
	**  first sort by user_rels[].sflag so that rels made before
	**  rf was cleared are at the bottom of the list
	*/

	if (ur1->sflag == ur2->sflag) {
		to_swap = (iwgt1 < iwgt2);
	}
	else {
		to_swap = (ur1->sflag > ur2->sflag);
	}

  return to_swap;
}


int compare_user_rels_relno(struct user_rj *ts1, struct user_rj *ts2) {
  /*
  **  Quick sort Comparison function for user rels. Sort key:
  **
  **  Used to sort by relj asc then weight desc.
  **  Now only sort by weight desc
  */
	
  register int to_swap = FALSE;

  to_swap = (ts1->relno > ts2->relno);

  return to_swap;
}

void sort_user_rels(struct user_rj *urels, int no_terms, int sort_flag) {
  /*
  **  sort_flag determines which compare_user_rels<n> function to use.
  **
  **    WGT_SORT (0) = compare_user_rels_wgt
  **  RELNO_SORT (1) = compare_user_rels_relno
  */

  if (sort_flag == WGT_SORT) {
    (void) qsort((char *) urels, no_terms,
								 sizeof(*urels), compare_user_rels_wgt);
  }
  else if (sort_flag == RELNO_SORT) {
    (void) qsort((char *) urels, no_terms,
								 sizeof(*urels), compare_user_rels_relno);
  }
  else {
    printf("Unknown sort flag. User rels not sorted.\n");
  }
}


