/*
**  weighting.c
**
**  Weighting functions - moved from utilities.c Apr 97
*/

#define G_EXTERN extern

#include "defs.h"

#undef MAX_WGT
#undef MIN_WGT

G_EXTERN void terminate_run(char *, char *, int);

int calc_little_r(int bss_set);
double calc_wgt(int np, int r, char term_type);
double calc_rsv(double weight, int r, char term_type);


int calc_little_r(int set) {
	int return_code;

  int little_r;
  int new_set;

	char bss_command[SMALL_BUFFER];
	char bss_result[SMALL_BUFFER];
  char *rbuf;
	
  sprintf(bss_command, "f s=%d s=%d op=and", set, BIGR_SET);

  if ((return_code = i0(bss_command, bss_result)) < 0) {
    terminate_run("calc_little_r()", bss_command, return_code);
  }
  
  rbuf = bss_result;
  sscanf(rbuf, "S%d np=%d ", &new_set, &little_r);
  
  sprintf(bss_command, "delete %d", new_set);
  if ((return_code = i0(bss_command, bss_result)) < 0) {
    terminate_run("calc_little_r()", bss_command, return_code);
  }
	
  return little_r;
}


double calc_wgt(int np, int r, char term_type) {
	int debugging = atoi(getenv("CALC_WGT_DEBUG"));

	int return_code;

  int rload = 0;
  int bigrload = 0;

  double weight;

	char bss_command[SMALL_BUFFER];
	char bss_result[SMALL_BUFFER];
  char *rbuf;
	
	if (debugging) {
		fprintf(stderr,
						"calc_wgt():: np = %d, r = %d, R = %d, term_type = [%c]\n",
						np, r, BIGR, toupper(term_type));
	}
	
  if (toupper(term_type) == USER_TERM) {
    rload = atoi(getenv("RLOAD"));
    bigrload = atoi(getenv("BIGRLOAD"));
  }

  /*
	**  Use little_r = 0 if it is negative.
	**  This will be the case for an extracted term that
	**  no longer exists because it only existed in 1 rel
	**  that has since been judged as non-rel.
	*/
	
  if (r < 0) {
    r = 0;
  }
	
  /*
  **  Calculate weight - now returned as a double.
  */

  sprintf(bss_command,
					"w fn=%d r=%d bigr=%d rload=%d bigrload=%d n=%d",
					atoi(getenv("WEIGHT_FUNCTION")),
					r, BIGR, rload, bigrload, np);
	
  if ((return_code = i0(bss_command, bss_result)) < 0) {
    terminate_run("calc_wgt()", bss_command, return_code);
  }

	weight = atof(bss_result);

	if (debugging) {
		fprintf(stderr,
						"calc_wgt:: rload=%d, bigrload=%d, weight = %10.3f\n",
						rload, bigrload, weight);
	}

  return weight;
}


double calc_rsv(double weight, int r, char term_type)
{
  /*
	**  U, (r+r')/(R+R')
	**  E, r/R
	*/

	int debugging = atoi(getenv("CALC_RSV_DEBUG"));

  int factor;
  int rload = 0;
  int bigrload = 0;

  double rsv;

	if (debugging) {
		fprintf(stderr,
						"calc_rsv():: weight = %.3f, r = %d, R = %d, term_type = [%c]\n",
						weight, r, BIGR, toupper(term_type));
	}
	
  if (toupper(term_type) == USER_TERM) {
    rload = r + atoi(getenv("RLOAD"));
    bigrload = BIGR + atoi(getenv("BIGRLOAD"));
  }
  else {
    rload = r;
    bigrload = BIGR;
  }

  rsv = weight * (double) rload / (double) bigrload;

	if (debugging) {
		fprintf(stderr,
						"calc_rsv():: rload=%d, bigrload=%d, rsv = %10.3f\n",
						rload, bigrload, rsv);
	}

  return rsv;
}

int MapWeight(double weight, double max_weight)
{
  /*
  **  This function maps the range [1..maxwt] for any given document set
  **  onto the range [MIN_WGT..MAX_WGT] for the hitlist and show
  **  displays so that the user will be given some indication of the
  **  relative weight of the document in the set.
  **
  **  The mapping function used is:
  **
  **    mapped_weight = max_value * weight / max_weight
  **
  **  where:
  **
  **    max_value = 1000.0
	**    max_weight is passed from bss_search()
  */

  int mapped_weight;
	double max_value = 1000.0;

	mapped_weight = round(max_value * weight / max_weight);

  if (mapped_weight > 1000) {
    mapped_weight = 1000;
  }
  else if (mapped_weight < 1) {
    mapped_weight = 0;
  }
	
  return mapped_weight;
}


void reweight(int no_terms, int final_search)
{
	/*
	**  final_search:  TRUE no relevance info - set little_r to 0
	**                FALSE relevance info - calc little_r for phrases
	*/

  int t;
	
  /*
	**  Only need to recalculate little_r for terms that were used
	**  to generate a document set since these were removed from
	**  the dcoument text before it was parsed to extract terms.
	**
	**  All other terms will have had there little_r values 
	**  correctly calculated in "do_extract()"
	*/
	
  for (t = 0; t < no_terms; t++) {
    /*
		**  Merged termsets:
		**
		**  Terms in an original termset but not in the extracted terms.
		**  All terms input by the user (including phrases) and any other
		**  terms in which there is a hyphen or space, need to have
		**  little_r recalculated.
		*/
		
    if (((termset + t)->term_type != EXTRACT_TERM) || (final_search)) {
      if ((termset + t)->bss_set != NONE_ASSIGNED) {
				/*
				**  i.e. a set has been generated - the term/phrase exists
				*/
				if (final_search) {
					(termset + t)->little_r = 0;
				}
				else {
					/*
					**  little_r will have been calculated correctly
					**  for single terms in do_extract(). Only need to
					**  calculate little_r for phrases.
					**
					**  Note this routine will also recalculate little_r
					**  for GSL phrases.
					*/
					
					if (strchr(PHRASE_OPS, toupper((termset + t)->operation)) != NULL) {
						/*
						**  A phrase
						*/
						(termset + t)->little_r = calc_little_r((termset + t)->bss_set);
					}
				}
      }
    }
		
    (termset + t)->weight = calc_wgt((termset + t)->np,
																 (termset + t)->little_r,
																 (termset + t)->term_type);
    (termset + t)->rsv = calc_rsv((termset + t)->weight,
															(termset + t)->little_r,
															(termset + t)->term_type);
  }
}

