##
##  TCL Interface: packing function (pack_it_all) in procedures.tcl
##
##  There is a bodge in here to stop the unwanted highlighting of
##  the contents of text widgets after double-clicking the mouse
##  in it. i.e. provide:
##
##    -selectbackground <text_widget_background_colour>
##    -selectborderwidth 0
##
##  as attributes of the widget.


proc gui {} {
	global intro_query_terms_title
	global TermEntryLabel_1
	global hitlist_title
	global rels_title
	global full_title
  global lr_value
  global rsv_value
	global qsize
	global rf_flag

	##  The global variable <rf_flag> is set when the program is
	##  first run by reading the FEEDBACK passed as a parameter
	##  when the program is run. 0 = RF OFF, 1 = RF ON.

	set lr_value 2
	set rsv_value 0.6
	set qsize 20
	
  . configure -bg SLateGray
  . configure -cursor left_ptr 

  set menu_name ".options_menu"

  wm title . $full_title

  ##  Main screen designing.
  ##
  ##  1. The term entry area.
  ##
  ##  1.1. The label

  frame .term_entry_header \
    -relief flat \
    -bg LightGrey \
    -bd 1 \
    -width 200

  ##  1.2.    The entry box
  ##  1.2.1.  Term entry label 

  label .term_entry_label \
    -relief flat \
    -width 200 \
    -bg LightGrey \
    -fg black \
    -font -*-helvetica-bold-r-normal--*-120-*-*-*-*-*-* \
    -text $TermEntryLabel_1

  ##  1.2.2.  The frame for the labels and the entry box

  frame .term_entry_frame \
    -relief flat \
    -bg LightGrey \
    -bd 1

  ##  1.2.3.  The term entry box
  ##
  ##  font was:
  ##
  ##    -font -*-helvetica-bold-r-normal--*-120-*-*-*-*-*-*
  ##
  ##  until NTD mucked up the font server!

  entry .term_entry_box \
    -relief groove \
    -width 200 \
    -bd 1 \
    -bg White \
    -fg Red \
		-selectbackground LightCyan \
    -font -*-fixed-bold-r-normal--15-140-*-*-*-*-*-* \
    -textvariable query

		bind .term_entry_box <Enter> {
				focus .term_entry_box
		}


  ##  2. Two frames to be packed below the term_entry_frame,
  ##     one on the left and one on the right.
  ##
  ##     In the lefthand frame will be packed the current query
  ##     and removed_terms areas.
  ##
  ##     In the righthand frame will be packed the hitlist and
  ##     positive rels areas.

  frame .left \
    -bg White \
    -relief flat

  frame .right \
    -bg White \
    -relief flat
  
  ##  3. The current query area - to go in .left
  ##
  ##  3.1. Main frame
  
  frame .current_query_frame \
    -relief groove \
		-bd 1 \
    -bg SlateGrey

  ##  3.2. Label

  label .current_query_label \
    -text $intro_query_terms_title \
    -relief groove \
    -bd 1 \
    -bg LightGrey \
    -fg Black \
    -font -*-fixed-bold-r-normal--15-140-*-*-*-*-*-*

  ##  3.3. Text widget

  text .current_query_text \
    -wrap "word" \
    -width 30 \
    -height 20 \
    -relief groove \
    -bd 1 \
    -bg White \
    -selectbackground white \
    -selectborderwidth 0 \
    -fg Black \
    -font -*-fixed-bold-r-normal--15-140-*-*-*-*-*-* \
		-yscrollcommand ".current_query_vscroll set"

##    -font -*-fixed-bold-r-normal--15-140-*-*-*-*-*-*

  ##  3.4. Vertical scroll bar

  scrollbar .current_query_vscroll \
    -command ".current_query_text yview" \
    -bg Lightgrey \
    -troughcolor SlateGrey \
    -highlightcolor LightGoldenRodYellow

  bind .current_query_text <Enter> {
    focus .current_query_vscroll
	}

  ##  4. The removed terms area - to go in .left,
  ##     below .current_query_frame
  ##
  ##  4.1. Main frame

  frame .removed_terms_frame \
    -relief groove \
    -bd 1 \
    -bg SlateGray

  ##  4.2. Text widget
	##  -wrap "word"
		
  
  text .removed_terms_text \
    -wrap "word" \
		-width 30 \
    -relief groove \
    -bd 1 \
    -bg Grey90 \
    -selectforeground Red \
    -selectbackground LightGrey \
    -selectborderwidth 0 \
    -fg Grey50 \
    -font -*-fixed-bold-r-normal--15-140-*-*-*-*-*-* \
    -yscrollcommand ".removed_terms_vscroll set"
  
  ##  4.3. Scroll bar

  scrollbar .removed_terms_vscroll \
    -command ".removed_terms_text yview" \
    -bg LightGrey \
    -troughcolor SlateGrey \
    -highlightcolor LightGoldenRodYellow

  ##  Click <Button-1> in this text widget to enable
  ##  explicit key bindings (e.g. up, down, next, prior).

  bind .removed_terms_text <Enter> {
    focus .removed_terms_vscroll
	}

  ##  5. The hitlist area - to go in .right
  ##
  ##  5.1. Main frame

  frame .hitlist_frame \
    -relief groove \
    -bd 1 \
    -bg SlateGray
  
  ##  5.2. Label

  label .hitlist_label \
    -text $hitlist_title \
    -relief groove \
    -bd 1 \
    -bg LightGrey \
    -fg Black \
    -font -*-fixed-bold-r-normal--15-140-*-*-*-*-*-*
  
  ##  5.3. Text widget

  text .hitlist_text \
    -width 55 \
    -height 20 \
    -relief groove \
    -bd 1 \
    -bg White \
    -fg Black \
    -selectbackground white \
    -selectborderwidth 0 \
    -font -*-fixed-bold-r-normal--15-140-*-*-*-*-*-* \
    -yscrollcommand ".hitlist_vscroll set"
  
  ##  5.4. Scroll bar

  scrollbar .hitlist_vscroll \
    -command ".hitlist_text yview" \
    -bg LightGrey \
    -troughcolor SlateGrey \
    -highlightcolor LightGoldenRodYellow

  ##  Move mouse into this text widget to enable explicit
	##  key bindings (e.g. up, down, next, prior).

  bind .hitlist_text <Enter> {
    focus .hitlist_vscroll
	}

  ##  6. The positive rels window - to go in .right,
  ##     below .hitlist_frame.
  ##
  ##  6.1. The main frame

  frame .positive_rels_frame \
    -relief groove \
    -bd 1 \
    -bg SlateGray
  
  ##  6.2. The text widget

  text .positive_rels_text \
    -width 55 \
    -relief groove \
    -bd 1 \
    -bg Grey90 \
    -fg Black \
    -font -*-fixed-bold-r-normal--15-140-*-*-*-*-*-* \
    -yscrollcommand ".positive_rels_vscroll set"
  
  ##  6.3. The scroll bar

  scrollbar .positive_rels_vscroll \
    -command ".positive_rels_text yview" \
    -bg LightGrey \
    -troughcolor SlateGrey \
    -highlightcolor LightGoldenRodYellow

  ##  Move mouse into this text widget to enable
  ##  explicit key bindings (e.g. up, down, next, prior).

  bind .positive_rels_text <Enter> {
    focus .positive_rels_vscroll
	}

  ## 7. The button area at the bottom of the screen, below .left & .right

  frame .main_buttons \
    -bg SlateGray \
    -relief groove \
    -bd 1

  button .main_buttons.search \
    -text "Search Database (^S)" \
    -command "db_search" \
    -relief groove \
    -bd 1 \
    -bg LightGrey \
    -activebackground LightGoldenRodYellow \
    -disabledforeground Grey60 \
    -state disabled

  button .main_buttons.options \
    -text "Query Options (^O)" \
    -command "post_menu $menu_name" \
    -relief groove \
    -bd 1 \
    -bg LightGrey \
    -activebackground LightGoldenRodYellow \
    -disabledforeground Grey60 \
    -state normal

   
  button .main_buttons.quit \
    -text "Exit Okapi (^X)" \
    -command "quit_search" \
    -relief groove \
    -bd 1 \
    -bg LightGrey \
    -activebackground LightGoldenRodYellow \
    -disabledforeground Grey60 \
    -state normal

  
  ##  8. Options popup menus

	set menu_name ".options_menu"
	
	menu $menu_name \
			-bg LightGrey \
			-activebackground LightGoldenRodYellow \
			-disabledforeground Grey \
			-bd 1 \
			-font -*-helvetica-medium-r-normal--*-140-*-*-*-*-*-* \
			-tearoff 0
  
	if { $rf_flag == 1 } {
		##  Clear relevance feedback menu option
	
		$menu_name add command \
				-label "  Clear Relevance Feedback  " \
				-command "clear_rf"
	}
	
	##  Clear current query
	##
	
	$menu_name add command \
      -label "  Clear Current Query  " \
      -command "clear_query"
	
	$menu_name add separator
	
	if { $rf_flag == 1 } {
		##  Set working query size menu option

		$menu_name add cascade \
				-label "  Set Working Query Size  " \
				-menu $menu_name.setsize
		
		menu $menu_name.setsize \
				-bg LightGrey \
				-fg black \
				-activebackground LightGoldenRodYellow \
				-activeforeground black \
				-disabledforeground grey \
				-bd 1 \
				-font -*-helvetica-medium-r-normal--*-140-*-*-*-*-*-* \
				-tearoff 0
		
		$menu_name.setsize add radio \
				-label "20 terms" \
				-variable qsize \
				-value 20 \
				-command "set_query_size 20 $menu_name"
		
		$menu_name.setsize add radio \
				-label "30 terms" \
				-variable qsize \
				-value 30 \
				-command "set_query_size 30 $menu_name"
		
		$menu_name.setsize add radio \
				-label "40 terms" \
				-variable qsize \
				-value 40 \
				-command "set_query_size 40 $menu_name"
		
		$menu_name add separator
	}
	
	if { $rf_flag == 1 } {
		##  Change LR Threshold menu option

		$menu_name add cascade \
				-label "  Change LR Threshold  " \
				-menu "$menu_name.lr_threshold"
		
		menu $menu_name.lr_threshold \
				-bg LightGrey \
				-fg black \
				-activebackground LightGoldenRodYellow \
				-activeforeground black \
				-disabledforeground grey \
				-bd 1 \
				-font -*-helvetica-medium-r-normal--*-140-*-*-*-*-*-* \
				-tearoff 0
		
		$menu_name.lr_threshold add radio \
				-label "1 or more doc" \
				-variable lr_value \
				-value 1 \
				-command "set_lr_threshold 1 $menu_name.lr_threshold"
		
		$menu_name.lr_threshold add radio \
				-label "2 or more docs" \
				-variable lr_value \
				-value 2 \
				-command "set_lr_threshold 2 $menu_name.lr_threshold"
		
		$menu_name.lr_threshold add radio \
				-label "3 or more docs" \
				-variable lr_value \
				-value 3 \
				-indicator 1 \
				-command "set_lr_threshold 3 $menu_name.lr_threshold"
		
		$menu_name.lr_threshold add radio \
				-label "4 or more docs" \
				-variable lr_value \
				-value 4 \
				-command "set_lr_threshold 4 $menu_name.lr_threshold"
		
		$menu_name add separator
		
		$menu_name add command \
				-label "  Fewer RF Terms  " \
				-command "adjust_rsv_factor U"
		
		$menu_name add command \
				-label "  More RF Terms  " \
				-command "adjust_rsv_factor D"
		
		$menu_name add separator
	}
	
	##  Cancel Menu
	##
	
	$menu_name add command \
      -label "  Cancel Menu  " \
      -command "cancel_menu $menu_name"

	bind . <Control-o> "post_menu $menu_name"
	bind . <Control-x> "quit_search"

  ##  End of menu definitions

  #####################
  ##  Start packing  ##
  #####################

  pack_it_all

  ##	disable_start

  #####################
  ##  End of packing ##
  #####################

  wm deiconify .

  ##  binding functions moved to procedures.tcl
  ##

##  focus default .term_entry_box
  focus -force .term_entry_box
  focus .term_entry_box
}
