/*Copyright (c) 2008, City University London
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and
/or other materials provided with the distribution.
    * Neither the name of the City University London nor the names of its contributors may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY D
IRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS O
F USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING N
EGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/******************************************************************************

  make_para_rec.c

  Extracted from make_parafile.c Nov 99 SW

  Only makes type 2 records (I don't see any point retaining other types)

  Record structure is:

  [fdnum (1-)
  offset of field from start of document
  field length
  number of paragraphs in field
  [offset from start of field of NEXT para
  sentence number (0-) of last sentence in PREV para]]

  where the whole thing repeats NF times and the last two fields
  number of paras in field times.

  fdnum is a single byte; all other fields are 1-4 byte encoded little-endian
  values (see w1to4() and r1to4())

  Returns length written or EOF

******************************************************************************/

#include "bss.h"
#include "para.h"


int
make_para_rec(int fdnum, 
	      char *fdstart, int fdlen,	/* Field text and length */
	      int toffset,	/* Offset of text in document */
	      FILE *parf,	/* Output files */
	      int maxtoks	/* Max tokens in a sentence */
	      )
{
  int paracount ;
  int paralen, l ;
  char *pp ;
  int se_num ;
  int sentence_count ;
  int outlen = 0 ;

  se_num = 0 ;
  putc(fdnum, parf) ; outlen++ ;
  outlen += w1to4(toffset, parf) ;
  outlen += w1to4(fdlen, parf) ;
  l = fdlen ; pp = fdstart ;
  paracount = count_paras(pp, l, INDENT|GAP) ;
  outlen += w1to4(paracount, parf) ;
  
  //printf("fdnum=%d, fdlen=%d, fdstart=%d, paracount=%d\n", fdnum,fdlen,fdstart,paracount);
  //printf("%s",pp);
  
  while ( l > 0 ) {
    paralen = find_para(pp, l, INDENT|GAP) ;
    /* Last arg temp, see the function (July 98) */
    sentence_count = count_sentences(pp, paralen, 0, maxtoks) ;
    se_num += sentence_count ;
    
    //printf("paralen=%d, sentence_count=%d\n%s", paralen, sentence_count,pp);
    pp += paralen ; l -= paralen;

    /* Then pp - 1 is last char in paragraph */
    /* Write out start of NEXT para, # of LAST sentence */
    outlen += w1to4(pp - fdstart, parf) ;
    outlen += w1to4(se_num - 1, parf) ;
  }
  if ( ferror(parf) ) outlen = EOF ;
  return outlen ;
}
