/*Copyright (c) 2008, City University London
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and
/or other materials provided with the distribution.
    * Neither the name of the City University London nor the names of its contributors may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY D
IRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS O
F USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING N
EGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/******************************************************************************

  mergeonly.c SW May 93

  Prog to do single merge pass on an acceptable number of Okapi index
  runfiles.

  Jan 99: generalized to accept any input file names (one can be stdin)

******************************************************************************/

#include "defines.h"

BOOL silent =  FALSE ;
u_int Secs ;
OFFSET_TYPE imerge(int num, FILE **, FILE *, int itype) ;

int
main(int argc, char *argv[])
{
  int numruns ;
  char **infile_pathnames ;
  FILE **infiles ;
  int itype ;
  OFFSET_TYPE result ;
  BOOL already_got_stdin = FALSE ;
  int i ;
  static char *usage = 
 "Usage: %s numruns input_filenames index_type\
One of the input filenames can be - for stdin.\
Writes output to stdout.\n" ;

  if ( argc < 4 || strcmp(argv[1], "-h") == 0 ) goto usg ;

  numruns = atoi(argv[1]) ;
  if ( numruns <= 0 || argc < numruns + 3 ) goto usg ;
  infile_pathnames = (char **)malloc(numruns * sizeof(char *)) ;
  infiles = (FILE **)malloc(numruns * sizeof(FILE *)) ;
  if ( infile_pathnames == NULL || infiles == NULL ) {
    fprintf(stderr, "Can't allocate memory for input filenames\n") ;
    exit(1) ;
  }
  for ( i = 0 ; i < numruns ; i++ ) {
    infile_pathnames[i] = argv[2 + i] ;
    if ( strcmp(infile_pathnames[i], "-") == 0 ) {
      if ( already_got_stdin ) {
	fprintf(stderr, "Can't have more than one input from stdin\n") ;
	goto usg ;
      }
      infiles[i] = stdin ;
      already_got_stdin = TRUE ;
    }
    else {
      infiles[i] = fopen(infile_pathnames[i], "rb") ;
      if ( infiles[i] == NULL ) {
	fprintf(stderr, "Can't open input file %s\n", infile_pathnames[i]) ;
	exit(1) ;
      }
    }
  }
  itype=atoi(argv[2 + numruns]) ;
  if ( itype < 4 || (itype > 13 && itype != 20) ) {
    fprintf(stderr, "Can only handle index types 4 to 14 and 20\n") ;
    exit(1) ;
  }
  result = imerge(numruns, infiles, stdout, itype) ;
  if ( result < 0 ) exit(1) ;
  else exit(0) ;

 usg:
  fprintf(stderr, usage, argv[0]) ;
  exit(1) ;
}


